# SOURCE FILENAME : orbiter.py
# AUTHOR          : Athul Pradeepkumar Girija, apradee@purdue.edu
# DATE CREATED    : 06/01/2022 07:51 AM
# DATE MODIFIED   : 06/01/2022 07:51 MT
# REMARKS         : Compute the coast trajectory from atmospheric exit until
#					apoapsis, the periapsis raise manuever dv, and the
#   				first full orbit post aerocapture. Also computes the
#					probe trajectory following a probe-targeting maneuver,
#					probe atmospheric entry conditions, and the orbiter trajectory
#					following an orbiter deflection manuever.


import numpy as np
from scipy import optimize
import matplotlib.pyplot as plt
from AMAT.planet import Planet
from AMAT.vehicle import Vehicle

from AMAT.approach import Approach

class Orbiter:
	"""
	Compute the coast trajectory from atmospheric exit until
	apoapsis, the periapsis raise manuever dv, and the
	first full orbit post aerocapture. Also computes the
	probe trajectory following a probe-targeting maneuver,
	probe atmospheric entry conditions, and the orbiter trajectory
	following an orbiter deflection manuever.

	Attributes
	----------
	vehicle : AMAT.vehicle.Vehicle
		Vehicle object which has been propagated until atmospheric exit
	peri_alt_km : float
		target orbit periapsis altitude, km
	r_periapsis_mag : float
		target orbit periapsis radial distance, m
	terminal_r : float
		planet-relative atmospheric exit state terminal radial distance, m
	terminal_theta : float
		planet-relative atmospheric exit state terminal longitude, rad
	terminal_phi : float
		planet-relative atmospheric exit state terminal latitude, rad
	terminal_v : float
		planet-relative atmospheric exit state terminal speed, m/s
	terminal_g : float
		planet-relative atmospheric exit state terminal flight-path angle, rad
	terminal_psi : float
		planet-relative atmospheric exit state terminal heading angle, rad
	terminal_r_vec : numpy.ndarray
		planet-relative atmospheric exit state radius vector, m
	terminal_r_hat_vec : numpy.ndarray
		planet-relative atmospheric exit state radius unit vector
	terminal_v_vec : numpy.ndarray
		planet-relative atmospheric exit state velocity vector, m/s
	terminal_r_hat_vec : numpy.ndarray
		planet-relative atmospheric exit state velocity unit vector
	terminal_v_ie_vec : numpy.ndarray
		inertial atmospheric exit state velocity vector, m/s
	terminal_v_ie_hat_vec : numpy.ndarray
		inertial atmospheric exit state velocity unit vector
	terminal_fpa_ie_deg : float
		inertial atmospheric exit state flight-path angle, deg
	terminal_fpa_ie_rad : float
		inertial atmospheric exit state flight-path angle, rad
	terminal_r_bi : float
		body-inertial atmospheric exit state terminal radial distance, m
	terminal_theta_bi : float
		body-inertial atmospheric exit state terminal longitude, rad
	terminal_phi_bi : float
		body-inertial atmospheric exit state terminal latitude, rad
	terminal_v_bi : float
		body-inertial atmospheric exit state terminal speed, m/s
	terminal_g_bi : float
		body-inertial atmospheric exit state terminal flight-path angle, rad
	terminal_psi_bi : float
		body-inertial atmospheric exit state terminal heading angle, rad
	terminal_E : float
		orbit total specific energy at atmospheric exit
	terminal_h : float
		orbit specific angular momentum scalar at atmospheric exit
	a : float
		initial capture orbit semi-major axis, m
	e : float
		initial capture orbit eccentricity
	i : float
		initial capture inclination, rad
	h : float
		initial capture specific angular momentum, SI units
	OMEGA : float
		initial capture RAAN, rad
	omega : float
		initial capture AoP, rad
	theta_star_exit : float
		true anomaly at atmospheric exit, rad
	x_coast_arr : numpy.ndarray
		body-inertial x-coordinate positions of coast trajectory from atmospheric exit until apoapsis,
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	y_coast_arr : numpy.ndarray
		body-inertial y-coordinate positions of coast trajectory from atmospheric exit until apoapsis,
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	z_coast_arr : numpy.ndarray
		body-inertial z-coordinate positions of coast trajectory from atmospheric exit until apoapsis,
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	r_apoapsis_vec : numpy.ndarray
		body-inertial radial vector of apoapsis, in terms of planet radius (vehicle.planetObj.RP)
	r_apoapsis_mag : float
		body-inertial radial apoapsis distance, m
	v_apoapsis_vec_coast : float
		body-inertial velocity vector at apoapsis for coast orbit, m/s
	v_apoapsis_mag_coast : float
		body-inertial velocity vector magnitude at apoapsis for coast orbit, m/s
	v_apoapsis_vec_orbit : numpy.ndarray
		body-inertial velocity vector at apoapsis for final orbit, m/s
	v_apoapsis_mag_orbit : float
		body-inertial velocity vector magnitude at apoapsis for final orbit, m/s
	PRM_dv_vec : numpy.ndarray
		periapsis raise manuever delta-V vector, m/s
	PRM_dv_mag : float
		periapsis raise manuever delta-V magnitude, m/s
	x_orbit_arr : numpy.ndarray
		body-inertial x-coordinate positions of one full final orbit after PRM
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	y_orbit_arr : numpy.ndarray
		body-inertial y-coordinate positions of one full final orbit after PRM
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	z_orbit_arr : numpy.ndarray
		body-inertial z-coordinate positions of one full final orbit after PRM
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	theta_star_probe_targeting : float
		true anomaly of the probe targeting manuever, from the initial capture orbit
		[0, 2*pi], rad
	a_probe_orbit : float
		probe approach orbit semi-major axis, m
	e_probe_orbit : float
		probe approach eccentricity
	i_probe_orbit : float
		probe approach inclination, rad
	h_probe_orbit : float
		probe approach  specific angular momentum, SI units
	OMEGA_probe_orbit  : float
		probe approach  RAAN, rad
	omega_probe_orbit : float
		probe approach  AoP, rad
	r_periapsis_probe : float
		periapsis radius of probe approach trajectory, m
	r_apoapsis_probe : float
		apoapsis radius of probe approach trajectory, m
	h_periapsis_probe : float
		periapsis altitude of probe approach trajectory, m
	h_EI : float
		probe atmospheric interface altitude, m
	r_EI : float
		probe atmospheric interface radius, m
	theta_star_probe_entry : float
		true anomaly at probe atmospheric entry, rad
	r_vec_bi_probe_entry : numpy.ndarray
		position vector at probe atmospheric entry, m
	r_vec_bi_probe_entry_unit : numpy.ndarray
		unit position vector at probe atmospheric entry
	r_vec_bi_probe_entry_mag : float
		position vector magnitude at probe atmospheric entry, m
	v_bi_vec_probe_entry : numpy.ndarray
		inertial velocity vector at probe atmospheric entry, m/s
	v_bi_mag_probe_entry : float
	 	inertial velocity magnitude at probe atmospheric entry, m/s
	v_bi_vec_probe_entry_unit : numpy.ndarray
		unit inertial velocity vector at probe atmospheric entry, m/s
	gamma_inertial_probe_entry : float
		inertial flight-path angle at probe atmospheric entry, rad
	latitude_probe_entry_bi : float
		inertial latitude at probe atmospheric entry, rad
	longitude_probe_entry_bi : float
		inertial longitude at probe atmospheric entry, rad
	v_vec_probe_entry_atm : numpy.ndarray
		atmosphere-relative velocity vector at probe atmospheric entry, m/s
	v_mag_probe_entry_atm : float
		atmosphere-relative velocity magnitude at probe atmospheric entry, m/s
	v_vec_probe_entry_atm_unit : numpy.ndarray
		atmosphere-relative velocity unit vector at probe atmospheric entry, m/s
	gamma_atm_probe_entry : float
		atmosphere-relative flight-path angle at probe atmospheric entry, rad
	v_atm_probe_entry_unit_proj : numpy.ndarray
		atmosphere-relative velocity unit vector projection on local horizontal plane
	atm_probe_entry_local_latitude_parallel_vec_unit : numpy.ndarray
		local parallel of latitude unit vector at probe atmospheric entry interface
	heading_atm_probe_entry : float
		atmosphere-relative heading angle at probe atmospheric entry, rad
	x_probe_arr : numpy.ndarray
		body-inertial x-coordinate positions of probe approach trajectory from
		probe targeting manuever until atmospheric entry
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	y_probe_arr : numpy.ndarray
		body-inertial y-coordinate positions of probe approach trajectory from
		probe targeting manuever until atmospheric entry
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	z_probe_arr : numpy.ndarray
		body-inertial z-coordinate positions of probe approach trajectory from
		probe targeting manuever until atmospheric entry
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	dV_mag_probe_targeting : float
		probe targeting maneuver delta-V magnitude, m/s
	dV_vec_probe_targeting : float
		probe targeting maneuver delta-V vector, m/s
	theta_star_orbiter_defl : float
		true anomaly of the orbiter deflection maneuever, from the probe targeting orbit
		[0, 2*pi], rad
	a_orbiter_defl : float
		orbiter deflection orbit semi-major axis, m
	e_orbiter_defl : float
		orbiter deflection orbit eccentricity
	i_orbiter_defl : float
		orbiter deflection orbit inclination, rad
	h_orbiter_defl : float
		orbiter deflection orbit  specific angular momentum, SI units
	OMEGA_orbiter_defl  : float
		orbiter deflection orbit  approach  RAAN, rad
	omega_orbiter_defl : float
		orbiter deflection  AoP, rad
	r_periapsis_orbiter_defl : float
		periapsis radius of orbiter deflection orbit trajectory, m
	r_apoapsis_orbiter_defl : float
		apoapsis radius of orbiter deflection orbit trajectory, m
	h_periapsis_orbiter_defl : float
		periapsis altitude of orbiter deflection orbit, m
	x_orbiter_defl_arr : numpy.ndarray
		body-inertial x-coordinate positions of orbiter deflection trajectory
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	y_orbiter_defl_arr : numpy.ndarray
		body-inertial y-coordinate positions of orbiter deflection trajectory
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	z_orbiter_defl_arr : numpy.ndarray
		body-inertial z-coordinate positions of orbiter deflection trajectory
		in non-dimensional units in terms of planet radius (vehicle.planetObj.RP)
	dV_mag_orbiter_defl : float
		orbiter deflection maneuver delta-V magnitude, m/s
	dV_vec_orbiter_defl : float
		orbiter deflection maneuver delta-V vector, m/s
	"""

	def __init__(self, vehicle, peri_alt_km):
		"""

		Parameters
		----------
		vehicle : AMAT.vehicle.Vehicle
			Vehicle object which has been propagated until atmospheric exit
		peri_alt_km : float
			orbit periapsis altitude, km
		"""

		self.vehicle = vehicle
		self.peri_alt_km = peri_alt_km
		self.r_periapsis_mag = self.peri_alt_km*1e3 + self.vehicle.planetObj.RP

		self.compute_exit_state_atm_relative()
		self.compute_exit_state_inertial()
		self.compute_coast_orbital_elements()
		self.compute_theta_star_exit()
		self.compute_coast_trajectory()
		self.compute_post_PRM_orbital_elements()
		self.compute_PRM_dv()
		self.compute_orbit_trajectory()

	def compute_exit_state_atm_relative(self):
		"""
		Extracts the exit state vector in atmosphere relative coordinates
		from the propagated vehicle exit state.
		"""

		self.terminal_r = self.vehicle.rc[self.vehicle.index - 1]
		self.terminal_theta = self.vehicle.thetac[self.vehicle.index - 1]
		self.terminal_phi = self.vehicle.phic[self.vehicle.index - 1]
		self.terminal_v = self.vehicle.vc[self.vehicle.index - 1]
		self.terminal_g = self.vehicle.gammac[self.vehicle.index - 1]
		self.terminal_psi = self.vehicle.psic[self.vehicle.index - 1]

	def compute_exit_state_inertial(self):
		"""
		Computes the exit state vector in body inertial frame.
		"""

		# Compute terminal radial vector
		self.terminal_r_vec = self.terminal_r * np.array([np.cos(self.terminal_phi) * np.cos(self.terminal_theta), \
												np.cos(self.terminal_phi) * np.sin(self.terminal_theta), np.sin(self.terminal_phi)])

		# Compute terminal radial unit vector
		self.terminal_r_hat_vec = self.terminal_r_vec / np.linalg.norm(self.terminal_r_vec)

		# Compute planet relative speed in Cartesian XYZ coordinates
		self.v_pr_x = self.terminal_v * np.sin(self.terminal_g) * np.cos(self.terminal_phi) * np.cos(self.terminal_theta) \
				 + self.terminal_v * np.cos(self.terminal_g) * np.cos(self.terminal_psi) * (-1 * np.sin(self.terminal_theta)) \
				 + self.terminal_v * np.cos(self.terminal_g) * np.sin(self.terminal_psi) * (-1 * np.sin(self.terminal_phi) * \
																			 np.cos(self.terminal_theta))

		self.v_pr_y = self.terminal_v * np.sin(self.terminal_g) * np.cos(self.terminal_phi) * np.sin(self.terminal_theta) + \
				 self.terminal_v * np.cos(self.terminal_g) * np.cos(self.terminal_psi) * np.cos(self.terminal_theta) + \
				 self.terminal_v * np.cos(self.terminal_g) * np.sin(self.terminal_psi) * (-1 * np.sin(self.terminal_phi) * \
																		   np.sin(self.terminal_theta))

		self.v_pr_z = self.terminal_v * np.sin(self.terminal_g) * np.sin(self.terminal_phi) + \
				 self.terminal_v * np.cos(self.terminal_g) * np.sin(self.terminal_psi) * np.cos(self.terminal_phi)

		self.terminal_v_vec = np.array([self.v_pr_x, self.v_pr_y, self.v_pr_z])
		self.terminal_v_hat_vec = self.terminal_v_vec / np.linalg.norm(self.terminal_v_vec)

		# Compute inertial speed in Cartesian XYZ coordinates
		self.v_ie_x = self.v_pr_x + self.terminal_r * self.vehicle.planetObj.OMEGA * np.cos(self.terminal_phi) * \
				 np.sin(self.terminal_theta) * (-1.0)

		self.v_ie_y = self.v_pr_y + self.terminal_r * self.vehicle.planetObj.OMEGA * np.cos(self.terminal_phi) * \
				 np.cos(self.terminal_theta)

		self.v_ie_z = self.v_pr_z

		# Compute inertial velocity magnitude
		self.v_ie_mag = np.sqrt(self.v_ie_x ** 2 + self.v_ie_y ** 2 + self.v_ie_z ** 2)

		# Compute inertial velocity vector
		self.terminal_v_ie_vec = np.array([self.v_ie_x, self.v_ie_y, self.v_ie_z])

		# Compute inertial velocity unit vector
		self.terminal_v_ie_hat_vec = self.terminal_v_ie_vec / np.linalg.norm(self.terminal_v_ie_vec)

		# Compute inertial flight path angle at exit using
		# terminal inertial radial and velocity vectors
		self.terminal_fpa_ie_deg = 90.0 - (180 / np.pi) * \
							  np.arccos(np.dot(self.terminal_r_hat_vec, self.terminal_v_ie_hat_vec))

		self.terminal_fpa_ie_rad = self.terminal_fpa_ie_deg * np.pi / 180.0

		self.terminal_r_bi = self.terminal_r
		self.terminal_theta_bi = self.terminal_theta
		self.terminal_phi_bi = self.terminal_phi
		self.terminal_v_bi = self.v_ie_mag
		self.terminal_g_bi = self.terminal_fpa_ie_rad
		self.terminal_psi_bi = self.terminal_psi

	def compute_coast_orbital_elements(self):

		"""
		Computes the orbital elements of the coast phase
		"""

		# Compute orbit energy using inertial speed
		self.terminal_E = self.vehicle.computeEnergyScalar(self.terminal_r, self.v_ie_mag)
		self.terminal_h = self.vehicle.computeAngMomScalar(self.terminal_r, self.v_ie_mag, self.terminal_fpa_ie_rad)

		# Compute semi-major axis and eccentricity of the post atmospheric exit orbit
		self.a = self.vehicle.computeSemiMajorAxisScalar(self.terminal_E)
		self.e = self.vehicle.computeEccScalar(self.terminal_h, self.terminal_E)
		self.h = np.sqrt(self.a * self.vehicle.planetObj.GM * (1 - self.e ** 2))

		self.h_vec_bi = np.cross(self.terminal_r_vec, self.terminal_v_ie_vec)
		self.h_vec_bi_unit = self.h_vec_bi / np.linalg.norm(self.h_vec_bi)

		self.i = np.arccos(self.h_vec_bi_unit[2])

		self.N_vec_bi = np.cross(np.array([0, 0, 1]), self.h_vec_bi)
		self.N_vec_bi_unit = self.N_vec_bi / np.linalg.norm(self.N_vec_bi)

		if self.N_vec_bi_unit[1] >= 0:
			self.OMEGA = np.arccos(self.N_vec_bi_unit[0])
		else:
			self.OMEGA = 2 * np.pi - np.arccos(self.N_vec_bi_unit[0])

		self.e_vec_bi = ((self.v_ie_mag**2 - self.vehicle.planetObj.GM/self.terminal_r)*self.terminal_r_vec - \
						 np.dot(self.terminal_r_vec, self.terminal_v_ie_vec)*self.terminal_v_ie_vec)/self.vehicle.planetObj.GM

		self.e_vec_bi_unit = self.e_vec_bi / np.linalg.norm(self.e_vec_bi)

		if self.e_vec_bi_unit[2] >= 0:
			self.omega = np.arccos(np.dot(self.N_vec_bi_unit, self.e_vec_bi_unit))
		else:
			self.omega = 2*np.pi - np.arccos(np.dot(self.N_vec_bi_unit, self.e_vec_bi_unit))


	def compute_theta_star_exit(self):
		self.theta_star_exit = np.arccos(((self.h ** 2 / (self.vehicle.planetObj.GM * self.terminal_r)) - 1) * (1.0 / self.e))

	def r_mag_bi(self, theta_star):
		"""
		Computes the position vector magnitude in body-inertial frame at a given true anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		r_mag_bi : float
			position vector magnitude, meters

		"""
		r_mag_bi = (self.h**2 / self.vehicle.planetObj.GM) / (1 + self.e*np.cos(theta_star))
		return r_mag_bi

	def pos_vec_bi(self, theta_star):
		"""
		Computes the position vector in body-inertial frame
		for a specified true-anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		pos_vec_bi : numpy.ndarray
			position vector in body-inertial frame, meters
		"""

		r = (self.h**2 / self.vehicle.planetObj.GM) / (1 + self.e*np.cos(theta_star))
		theta = theta_star + self.omega

		rx_unit = np.cos(self.OMEGA)*np.cos(theta) - np.sin(self.OMEGA)*np.cos(self.i)*np.sin(theta)
		ry_unit = np.sin(self.OMEGA)*np.cos(theta) + np.cos(self.OMEGA)*np.cos(self.i)*np.sin(theta)
		rz_unit = np.sin(self.i)*np.sin(theta)

		pos_vec_bi = r*np.array([rx_unit, ry_unit, rz_unit])
		return pos_vec_bi

	def pos_vec_bi_probe(self, theta_star):
		"""
		Computes the position vector in body-inertial frame
		of the probe trajectory for a specified true-anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		pos_vec_bi : numpy.ndarray
			position vector in body-inertial frame, meters
		"""

		r = (self.h_probe_orbit**2 / self.vehicle.planetObj.GM) / (1 + self.e_probe_orbit*np.cos(theta_star))
		theta = theta_star + self.omega

		rx_unit = np.cos(self.OMEGA_probe_orbit)*np.cos(theta) - np.sin(self.OMEGA_probe_orbit)*np.cos(self.i)*np.sin(theta)
		ry_unit = np.sin(self.OMEGA_probe_orbit)*np.cos(theta) + np.cos(self.OMEGA_probe_orbit)*np.cos(self.i)*np.sin(theta)
		rz_unit = np.sin(self.i_probe_orbit)*np.sin(theta)

		pos_vec_bi = r*np.array([rx_unit, ry_unit, rz_unit])
		return pos_vec_bi

	def pos_vec_bi_orbiter_defl(self, theta_star):
		"""
		Computes the position vector in body-inertial frame
		of the orbiter deflection trajectory for a specified true-anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		pos_vec_bi : numpy.ndarray
			position vector in body-inertial frame, meters
		"""

		r = (self.h_orbiter_defl**2 / self.vehicle.planetObj.GM) / (1 + self.e_orbiter_defl*np.cos(theta_star))
		theta = theta_star + self.omega

		rx_unit = np.cos(self.OMEGA_orbiter_defl)*np.cos(theta) - np.sin(self.OMEGA_orbiter_defl)*np.cos(self.i_orbiter_defl)*np.sin(theta)
		ry_unit = np.sin(self.OMEGA_orbiter_defl)*np.cos(theta) + np.cos(self.OMEGA_orbiter_defl)*np.cos(self.i_orbiter_defl)*np.sin(theta)
		rz_unit = np.sin(self.i_orbiter_defl)*np.sin(theta)

		pos_vec_bi = r*np.array([rx_unit, ry_unit, rz_unit])
		return pos_vec_bi

	def vel_vec_bi(self, theta_star):
		"""
		Computes the velocity vector in body-inertial frame at a given true anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		vel_vec_bi : numpy.ndarray
			velocity vector in body-inertial frame, m/s

		"""

		r_mag_bi = self.r_mag_bi(theta_star)
		v_mag_bi = np.sqrt(self.vehicle.planetObj.GM*(2/r_mag_bi - 1/self.a))

		gamma = 1*np.arccos(self.h/(r_mag_bi*v_mag_bi*1.0000001))

		vr = v_mag_bi*np.sin(gamma)
		vt = v_mag_bi*np.cos(gamma)
		theta = theta_star + self.omega

		vx = vr*( np.cos(theta)*np.cos(self.OMEGA) - np.sin(theta)*np.cos(self.i)*np.sin(self.OMEGA)) +\
		     vt*(-np.sin(theta)*np.cos(self.OMEGA) - np.cos(theta)*np.cos(self.i)*np.sin(self.OMEGA))

		vy = vr*( np.cos(theta)*np.sin(self.OMEGA) + np.sin(theta)*np.cos(self.i)*np.cos(self.OMEGA)) +\
		     vt*( np.cos(theta)*np.cos(self.i)*np.cos(self.OMEGA) - np.sin(theta)*np.sin(self.OMEGA))

		vz = vr*np.sin(theta)*np.sin(self.i) + vt*np.cos(theta)*np.sin(self.i)

		vel_vec_bi = np.array([vx, vy, vz])
		return vel_vec_bi


	def vel_vec_bi_probe(self, theta_star):
		"""
		Computes the velocity vector  in body-inertial frame for the probe trajectory
		at a given true anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		vel_vec_bi : numpy.ndarray
			velocity vector in body-inertial frame, m/s

		"""

		r_mag_bi = np.linalg.norm(self.pos_vec_bi_probe(theta_star))
		v_mag_bi = np.sqrt(self.vehicle.planetObj.GM*(2/r_mag_bi - 1/self.a_probe_orbit))

		gamma = -1*np.arccos(self.h_probe_orbit/(r_mag_bi*v_mag_bi*1.0000001))

		vr = v_mag_bi*np.sin(gamma)
		vt = v_mag_bi*np.cos(gamma)
		theta = theta_star + self.omega_probe_orbit

		vx = vr*( np.cos(theta)*np.cos(self.OMEGA_probe_orbit) - np.sin(theta)*np.cos(self.i_probe_orbit)*np.sin(self.OMEGA_probe_orbit)) +\
		     vt*(-np.sin(theta)*np.cos(self.OMEGA_probe_orbit) - np.cos(theta)*np.cos(self.i_probe_orbit)*np.sin(self.OMEGA_probe_orbit))

		vy = vr*( np.cos(theta)*np.sin(self.OMEGA_probe_orbit) + np.sin(theta)*np.cos(self.i_probe_orbit)*np.cos(self.OMEGA_probe_orbit)) +\
		     vt*( np.cos(theta)*np.cos(self.i_probe_orbit)*np.cos(self.OMEGA_probe_orbit) - np.sin(theta)*np.sin(self.OMEGA_probe_orbit))

		vz = vr*np.sin(theta)*np.sin(self.i_probe_orbit) + vt*np.cos(theta)*np.sin(self.i_probe_orbit)

		vel_vec_bi = np.array([vx, vy, vz])
		return vel_vec_bi


	def compute_coast_trajectory(self, num_points=301):
		"""
		Computes the coast trajectory from atmospheric exit until first apoapsis

		Parameters
		----------
		num_points : int
			number of grid points

		Returns
		-------

		"""

		theta_star_arr = np.linspace(self.theta_star_exit, np.pi, num_points)
		pos_vec_bi_arr = self.pos_vec_bi(theta_star_arr) / self.vehicle.planetObj.RP

		self.x_coast_arr = pos_vec_bi_arr[0][:]
		self.y_coast_arr = pos_vec_bi_arr[1][:]
		self.z_coast_arr = pos_vec_bi_arr[2][:]

		self.r_apoapsis_vec =  np.array([self.x_coast_arr[-1],
										self.y_coast_arr[-1],
										self.z_coast_arr[-1]])

		self.r_apoapsis_mag = np.linalg.norm(self.r_apoapsis_vec)*self.vehicle.planetObj.RP

		self.v_apoapsis_vec_coast = self.vel_vec_bi(np.pi)
		self.v_apoapsis_mag_coast = np.linalg.norm(self.v_apoapsis_vec_coast)

	def compute_post_PRM_orbital_elements(self):
		"""
		Computes the orbit elements for orbit with periapsis raised
		"""

		self.e = (self.r_apoapsis_mag - self.r_periapsis_mag) / (self.r_apoapsis_mag + self.r_periapsis_mag)
		self.a = self.r_apoapsis_mag / (1 + self.e)
		self.h = np.sqrt(self.a * self.vehicle.planetObj.GM * (1 - self.e ** 2))

	def compute_PRM_dv(self):
		"""
		Computes the periapsis raise manuever delta-V vector and magnitude
		"""

		self.v_apoapsis_vec_orbit = self.vel_vec_bi(np.pi)
		self.v_apoapsis_mag_orbit = np.linalg.norm(self.v_apoapsis_vec_orbit)

		self.PRM_dv_vec = self.v_apoapsis_vec_orbit - self.v_apoapsis_vec_coast
		self.PRM_dv_mag = np.linalg.norm(self.PRM_dv_vec)

	def compute_orbit_trajectory(self, num_points=601):
		"""
		Computes the coordinates of initial capture orbit after periapsis raise manuever

		Parameters
		----------
		num_points : int
			grid points for full final orbit trajectory


		"""
		theta_star_arr = np.linspace(np.pi, 3*np.pi, num_points)
		pos_vec_bi_arr = self.pos_vec_bi(theta_star_arr) / self.vehicle.planetObj.RP

		self.x_orbit_arr = pos_vec_bi_arr[0][:]
		self.y_orbit_arr = pos_vec_bi_arr[1][:]
		self.z_orbit_arr = pos_vec_bi_arr[2][:]

	def vel_vec_probe_entry_atm(self):
		"""
		Computes the atmosphere-relative velocity vector magnitude in body-inertial frame
		of the probe at atmospheric entry interface.

		Parameters
		----------

		Returns
		-------
		vel_vec_entry_atm : numpy.ndarray
			atmosphere-relative velocity vector in body-inertial frame, at
			atmospheric entry interface, m/s

		"""
		v_entry_atm_x = self.v_bi_vec_probe_entry[0] - \
		          self.r_vec_bi_probe_entry_mag * self.vehicle.planetObj.OMEGA * \
		          np.cos(self.latitude_probe_entry_bi) * (-np.sin(self.longitude_probe_entry_bi))
		v_entry_atm_y = self.v_bi_vec_probe_entry[1] - \
		          self.r_vec_bi_probe_entry_mag * self.vehicle.planetObj.OMEGA * \
		          np.cos(self.latitude_probe_entry_bi) * (np.cos(self.longitude_probe_entry_bi))
		v_entry_atm_z = self.v_bi_vec_probe_entry[2]

		vel_vec_entry_atm = np.array([v_entry_atm_x, v_entry_atm_y, v_entry_atm_z])
		return vel_vec_entry_atm


	def compute_probe_targeting_trajectory(self, theta_star, dV, num_points=301):
		"""
		Computes the coordinates of probe approach trajectory following a probe targeting manuever
		at theta_star of the initial capture orbit.

		Parameters
		----------
		theta_star : float
			true anomaly of the initial capture orbit at which the probe targeting maneuever
			is performed [0,2pi], rad
		dV : float
			delta-V magnitude of the probe targeting maneuver, along the direction
			of orbital velocity, can be positive or negative. Must be negative to
			deliver probe into the atmosphere from the initial capture orbit. m/s
		num_points : int
			number of grid points for the probe approach trajectory
		"""

		self.theta_star_probe_targeting = theta_star
		self.dV_mag_probe_targeting = dV

		# Compute state vector (r1,v1) at theta_star before manuever
		r_vec_1 = self.pos_vec_bi(theta_star)
		v_vec_1 = self.vel_vec_bi(theta_star)

		# Compute state vector (r2, v2) after applying dV
		r_vec_2 = r_vec_1
		v_vec_2 = v_vec_1 + dV*v_vec_1/(np.linalg.norm(v_vec_1))

		self.dV_vec_probe_targeting =  dV*v_vec_1/(np.linalg.norm(v_vec_1))

		r2 = np.linalg.norm(r_vec_2)
		v2 = np.linalg.norm(v_vec_2)

		r_vec_2_unit = r_vec_2 / np.linalg.norm(r_vec_2)
		v_vec_2_unit = v_vec_2 / np.linalg.norm(v_vec_2)

		ie_fpa_2_deg     = 90.0 - (180 / np.pi) * \
								   np.arccos(np.dot(r_vec_2_unit, v_vec_2_unit))
		ie_fpa_2_rad     = ie_fpa_2_deg*np.pi/180

		# Compute orbit energy and angular momentum post manueuver
		self.E_post_maneuver = self.vehicle.computeEnergyScalar(r2, v2)
		self.h_post_maneuver = self.vehicle.computeAngMomScalar(r2, v2, ie_fpa_2_rad)

		# Compute semi-major axis and eccentricity of the probe orbit
		self.a_probe_orbit = self.vehicle.computeSemiMajorAxisScalar(self.E_post_maneuver)
		self.e_probe_orbit = self.vehicle.computeEccScalar(self.h_post_maneuver, self.E_post_maneuver)
		self.h_probe_orbit = np.sqrt(self.a_probe_orbit * self.vehicle.planetObj.GM * (1 - self.e_probe_orbit ** 2))

		self.h_vec_bi_probe_orbit = np.cross(r_vec_2, v_vec_2)
		self.h_vec_bi_unit_probe_orbit = self.h_vec_bi_probe_orbit / np.linalg.norm(self.h_vec_bi_probe_orbit)

		self.i_probe_orbit = np.arccos(self.h_vec_bi_unit_probe_orbit[2])

		self.N_vec_bi_probe_orbit = np.cross(np.array([0, 0, 1]), self.h_vec_bi_probe_orbit)
		self.N_vec_bi_unit_probe_orbit = self.N_vec_bi_probe_orbit / np.linalg.norm(self.N_vec_bi_probe_orbit)

		if self.N_vec_bi_unit_probe_orbit[1] >= 0:
			self.OMEGA_probe_orbit = np.arccos(self.N_vec_bi_unit_probe_orbit[0])
		else:
			self.OMEGA_probe_orbit = 2 * np.pi - np.arccos(self.N_vec_bi_unit_probe_orbit[0])

		self.e_vec_bi_probe_orbit = ((v2 ** 2 - self.vehicle.planetObj.GM / r2) * r_vec_2 - \
						 			np.dot(r_vec_2, v_vec_2) * v_vec_2) / self.vehicle.planetObj.GM

		self.e_vec_bi_unit_probe_orbit = self.e_vec_bi_probe_orbit / np.linalg.norm(self.e_vec_bi_probe_orbit)

		if self.e_vec_bi_unit_probe_orbit[2] >= 0:
			self.omega_probe_orbit = np.arccos(np.dot(self.N_vec_bi_unit_probe_orbit, self.e_vec_bi_unit_probe_orbit))
		else:
			self.omega_probe_orbit = 2 * np.pi - np.arccos(np.dot(self.N_vec_bi_unit_probe_orbit, self.e_vec_bi_unit_probe_orbit))

		self.r_periapsis_probe = self.a_probe_orbit * (1 - self.e_probe_orbit)
		self.r_apoapsis_probe = self.a_probe_orbit * (1 + self.e_probe_orbit)
		self.h_periapsis_probe = self.r_periapsis_probe - self.vehicle.planetObj.RP

		if self.h_periapsis_probe < self.vehicle.planetObj.h_skip:
			self.h_EI = self.vehicle.planetObj.h_skip
			self.r_EI = self.vehicle.planetObj.RP + self.h_EI
			self.theta_star_probe_entry = 2*np.pi - np.arccos(((self.h_probe_orbit ** 2 / (self.vehicle.planetObj.GM * self.r_EI)) - 1) * (1.0 / self.e_probe_orbit))

			self.r_vec_bi_probe_entry = self.pos_vec_bi_probe(self.theta_star_probe_entry)
			self.r_vec_bi_probe_entry_unit = self.r_vec_bi_probe_entry / np.linalg.norm(self.r_vec_bi_probe_entry)
			self.r_vec_bi_probe_entry_mag = np.linalg.norm(self.r_vec_bi_probe_entry)

			self.v_bi_vec_probe_entry = self.vel_vec_bi_probe(self.theta_star_probe_entry)
			self.v_bi_mag_probe_entry = np.linalg.norm(self.v_bi_vec_probe_entry)
			self.v_bi_vec_probe_entry_unit = self.v_bi_vec_probe_entry / self.v_bi_mag_probe_entry

			self.gamma_inertial_probe_entry = -1 * np.arccos(self.h_probe_orbit / (self.r_EI * self.v_bi_mag_probe_entry))

			self.latitude_probe_entry_bi = np.arcsin(self.r_vec_bi_probe_entry_unit[2])
			self.longitude_probe_entry_bi = np.arctan(self.r_vec_bi_probe_entry_unit[1] / self.r_vec_bi_probe_entry_unit[0])

			self.v_vec_probe_entry_atm = self.vel_vec_probe_entry_atm()
			self.v_mag_probe_entry_atm = np.linalg.norm(self.v_vec_probe_entry_atm)
			self.v_vec_probe_entry_atm_unit = self.v_vec_probe_entry_atm / self.v_mag_probe_entry_atm

			self.gamma_atm_probe_entry = np.pi / 2 - np.arccos(np.dot(self.r_vec_bi_probe_entry_unit, self.v_bi_vec_probe_entry_unit))

			self.v_atm_probe_entry_unit_proj = self.v_vec_probe_entry_atm_unit  - \
											 np.dot(self.v_vec_probe_entry_atm_unit ,
													self.r_vec_bi_probe_entry_unit) * self.r_vec_bi_probe_entry_unit

			self.atm_probe_entry_local_latitude_parallel_vec_unit = np.array([-np.sin(self.longitude_probe_entry_bi),
															  				   np.cos(self.longitude_probe_entry_bi),
															  				   0.0])

			self.heading_atm_probe_entry = np.arccos(np.dot(self.v_atm_probe_entry_unit_proj,
													  self.atm_probe_entry_local_latitude_parallel_vec_unit))

		else:
			print("Probe orbit periapsis outside atmospheric skip altitude.")

		theta_star_arr = np.linspace(theta_star, self.theta_star_probe_entry, num_points)
		pos_vec_bi_arr = self.pos_vec_bi_probe(theta_star_arr) / self.vehicle.planetObj.RP

		self.x_probe_orbit_arr = pos_vec_bi_arr[0][:]
		self.y_probe_orbit_arr = pos_vec_bi_arr[1][:]
		self.z_probe_orbit_arr = pos_vec_bi_arr[2][:]

	def compute_orbiter_deflection_trajectory(self, theta_star, dV, num_points=601):
		"""
		Computes the coordinates of orbiter deflection trajectory following an orbiter deflection manuever
		at theta_star of the probe approach trajectory orbit.

		Parameters
		----------
		theta_star : float
			true anomaly of the probe approach trajectory orbit at which the orbiter deflection manuever
			is performed [0,2pi], rad
		dV : float
			delta-V magnitude of the orbiter deflection manuever, along the direction
			of orbital velocity, can be positive or negative. Must be positive to
			raise the periapsis outside the atmosphere. m/s
		num_points : int
			number of grid points for the orbiter deflection trajectory
		"""

		self.theta_star_orbiter_defl = theta_star
		self.dV_mag_orbiter_deflection = dV


		# Compute state vector (r1,v1) at theta_star before manuever
		r_vec_1 = self.pos_vec_bi_probe(theta_star)
		v_vec_1 = self.vel_vec_bi_probe(theta_star)

		# Compute state vector (r2, v2) after applying dV
		r_vec_2 = r_vec_1
		v_vec_2 = v_vec_1 + dV * v_vec_1 / (np.linalg.norm(v_vec_1))

		self.dV_vec_orbiter_deflection =  dV * v_vec_1 / (np.linalg.norm(v_vec_1))

		r2 = np.linalg.norm(r_vec_2)
		v2 = np.linalg.norm(v_vec_2)

		r_vec_2_unit = r_vec_2 / np.linalg.norm(r_vec_2)
		v_vec_2_unit = v_vec_2 / np.linalg.norm(v_vec_2)

		ie_fpa_2_deg = 90.0 - (180 / np.pi) * \
					   np.arccos(np.dot(r_vec_2_unit, v_vec_2_unit))
		ie_fpa_2_rad = ie_fpa_2_deg * np.pi / 180

		# Compute orbit energy and angular momentum for orbiter deflection trajectory
		self.E_orbiter_defl = self.vehicle.computeEnergyScalar(r2, v2)
		self.h_orbiter_defl = self.vehicle.computeAngMomScalar(r2, v2, ie_fpa_2_rad)

		# Compute semi-major axis and eccentricity of the probe orbit
		self.a_orbiter_defl = self.vehicle.computeSemiMajorAxisScalar(self.E_orbiter_defl)
		self.e_orbiter_defl = self.vehicle.computeEccScalar(self.h_orbiter_defl, self.E_orbiter_defl)
		self.h_orbiter_defl = np.sqrt(self.a_orbiter_defl * self.vehicle.planetObj.GM * (1 - self.e_orbiter_defl ** 2))

		self.h_vec_bi_orbiter_defl  = np.cross(r_vec_2, v_vec_2)
		self.h_vec_bi_unit_orbiter_defl  = self.h_vec_bi_orbiter_defl  / np.linalg.norm(self.h_vec_bi_orbiter_defl)

		self.i_orbiter_defl = np.arccos(self.h_vec_bi_unit_orbiter_defl[2])

		self.N_vec_bi_orbiter_defl = np.cross(np.array([0, 0, 1]), self.h_vec_bi_orbiter_defl)
		self.N_vec_bi_unit_orbiter_defl = self.N_vec_bi_orbiter_defl / np.linalg.norm(self.N_vec_bi_orbiter_defl)

		if self.N_vec_bi_unit_orbiter_defl[1] >= 0:
			self.OMEGA_orbiter_defl = np.arccos(self.N_vec_bi_unit_orbiter_defl[0])
		else:
			self.OMEGA_orbiter_defl = 2 * np.pi - np.arccos(self.N_vec_bi_unit_orbiter_defl[0])

		self.e_vec_bi_orbiter_defl = ((v2 ** 2 - self.vehicle.planetObj.GM / r2) * r_vec_2 -
									 np.dot(r_vec_2, v_vec_2) * v_vec_2) / self.vehicle.planetObj.GM

		self.e_vec_bi_unit_orbiter_defl =  self.e_vec_bi_orbiter_defl  / np.linalg.norm(self.e_vec_bi_orbiter_defl)

		if self.e_vec_bi_unit_orbiter_defl[2] >= 0:
			self.omega_orbiter_defl = np.arccos(np.dot(self.N_vec_bi_unit_orbiter_defl, self.e_vec_bi_unit_orbiter_defl))
		else:
			self.omega_orbiter_defl = 2 * np.pi - np.arccos(np.dot(self.N_vec_bi_unit_orbiter_defl, self.e_vec_bi_unit_orbiter_defl))

		self.r_periapsis_orbiter_defl = self.a_orbiter_defl * (1 - self.e_orbiter_defl)
		self.r_apoapsis_orbiter_defl = self.a_orbiter_defl* (1 + self.e_orbiter_defl)
		self.h_periapsis_orbiter_defl = self.r_periapsis_orbiter_defl - self.vehicle.planetObj.RP

		theta_star_arr = np.linspace(0, 2*np.pi, num_points)
		pos_vec_bi_arr = self.pos_vec_bi_orbiter_defl(theta_star_arr) / self.vehicle.planetObj.RP

		self.x_orbiter_defl_arr = pos_vec_bi_arr[0][:]
		self.y_orbiter_defl_arr = pos_vec_bi_arr[1][:]
		self.z_orbiter_defl_arr = pos_vec_bi_arr[2][:]

		if self.h_periapsis_orbiter_defl < self.vehicle.planetObj.h_skip:
			print("Periapsis inside atmospheric skip altitude.")


class PropulsiveOrbiter:
	"""
	Compute the orbit of a propulsive insertion orbiter after OI burn.
	"""

	def __init__(self, approach, apoapsis_alt_km):
		"""

		Parameters
		----------
		approach : AMAT.approach.Approach
			Approach object which has been propagated until periapsis
		apoapsis_alt_km : float
			orbit apoapsis altitude, km
		"""

		self.approach = approach
		self.apoapsis_alt_km = apoapsis_alt_km

		self.approach_terminal_r = self.approach.r_vec_rp_bi
		self.approach_terminal_v = self.approach.v_vec_rp_bi

		self.rp = self.approach.rp
		self.ra = self.approach.planetObj.RP + self.apoapsis_alt_km*1e3

		self.a = (self.rp + self.ra)/2
		self.e = (self.ra - self.rp) / (self.ra + self.rp)
		self.h = np.sqrt(self.a * self.approach.planetObj.GM * (1 - self.e ** 2))

		self.i = self.approach.i
		self.OMEGA = self.approach.OMEGA
		self.omega = self.approach.omega

		self.vp_vec = self.vel_vec_bi(theta_star=0)
		self.vp = np.linalg.norm(self.vp_vec)

		self.DV_OI_vec = self.approach_terminal_v - self.vp_vec
		self.DV_OI_mag = np.linalg.norm(self.DV_OI_vec)

		self.vp_check = np.sqrt(self.approach.planetObj.GM * (2 / self.rp - 1 / self.a))
		self.DV_OI_mag_check = np.linalg.norm(self.approach.v_vec_rp_bi) - self.vp_check

		self.compute_orbit_trajectory()


	def r_mag_bi(self, theta_star):
		"""
		Computes the position vector magnitude in body-inertial frame at a given true anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		r_mag_bi : float
			position vector magnitude, meters

		"""
		r_mag_bi = (self.h**2 / self.approach.planetObj.GM) / (1 + self.e*np.cos(theta_star))
		return r_mag_bi

	def pos_vec_bi(self, theta_star):
		"""
		Computes the position vector in body-inertial frame
		for a specified true-anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		pos_vec_bi : numpy.ndarray
			position vector in body-inertial frame, meters
		"""

		r = (self.h**2 / self.approach.planetObj.GM) / (1 + self.e*np.cos(theta_star))
		theta = theta_star + self.omega

		rx_unit = np.cos(self.OMEGA)*np.cos(theta) - np.sin(self.OMEGA)*np.cos(self.i)*np.sin(theta)
		ry_unit = np.sin(self.OMEGA)*np.cos(theta) + np.cos(self.OMEGA)*np.cos(self.i)*np.sin(theta)
		rz_unit = np.sin(self.i)*np.sin(theta)

		pos_vec_bi = r*np.array([rx_unit, ry_unit, rz_unit])
		return pos_vec_bi

	def vel_vec_bi(self, theta_star):
		"""
		Computes the velocity vector in body-inertial frame at a given true anomaly.

		Parameters
		----------
		theta_star : float
			true anomaly, rad

		Returns
		-------
		vel_vec_bi : numpy.ndarray
			velocity vector in body-inertial frame, m/s

		"""

		r_mag_bi = self.r_mag_bi(theta_star)
		v_mag_bi = np.sqrt(self.approach.planetObj.GM*(2/r_mag_bi - 1/self.a))

		gamma = 1*np.arccos(self.h/(r_mag_bi*v_mag_bi*1.0000001))

		vr = v_mag_bi*np.sin(gamma)
		vt = v_mag_bi*np.cos(gamma)
		theta = theta_star + self.omega

		vx = vr*( np.cos(theta)*np.cos(self.OMEGA) - np.sin(theta)*np.cos(self.i)*np.sin(self.OMEGA)) +\
		     vt*(-np.sin(theta)*np.cos(self.OMEGA) - np.cos(theta)*np.cos(self.i)*np.sin(self.OMEGA))

		vy = vr*( np.cos(theta)*np.sin(self.OMEGA) + np.sin(theta)*np.cos(self.i)*np.cos(self.OMEGA)) +\
		     vt*( np.cos(theta)*np.cos(self.i)*np.cos(self.OMEGA) - np.sin(theta)*np.sin(self.OMEGA))

		vz = vr*np.sin(theta)*np.sin(self.i) + vt*np.cos(theta)*np.sin(self.i)

		vel_vec_bi = np.array([vx, vy, vz])
		return vel_vec_bi

	def compute_orbit_trajectory(self, num_points=601):
		"""
		Computes the coordinates of initial capture orbit of the propulsive orbiter

		Parameters
		----------
		num_points : int
			grid points for full final orbit trajectory


		"""
		theta_star_arr = np.linspace(0, 2*np.pi, num_points)
		pos_vec_bi_arr = self.pos_vec_bi(theta_star_arr) / self.approach.planetObj.RP

		self.x_orbit_arr = pos_vec_bi_arr[0][:]
		self.y_orbit_arr = pos_vec_bi_arr[1][:]
		self.z_orbit_arr = pos_vec_bi_arr[2][:]

	def compute_timed_orbit_trajectory(self, t_seconds, num_points=101):
		"""
		Computes the coordinates of initial capture orbit of the propulsive orbiter
		as a function of time.

		Parameters
		----------
		t_seconds : float
			time to compute the orbit trajectory from periapsis, seconds
		num_points : int
			number of grid points in the time interval
		"""

		self.P = 2 * np.pi * np.sqrt(self.a ** 3 / self.approach.planetObj.GM)

		self.t_array = np.linspace(0, t_seconds, num_points)
		self.M_array = (2*np.pi*self.t_array)/self.P

		self.E_array = np.array([])
		self.theta_star_array = np.array([])

		for M in self.M_array:
			E = self.compute_E(M)
			theta_star = self.compute_theta_star(E)

			self.E_array = np.append(self.E_array, E)
			self.theta_star_array = np.append(self.theta_star_array, theta_star)

		pos_vec_bi_arr = self.pos_vec_bi(self.theta_star_array) / self.approach.planetObj.RP

		self.x_timed_orbit_arr = pos_vec_bi_arr[0][:]
		self.y_timed_orbit_arr = pos_vec_bi_arr[1][:]
		self.z_timed_orbit_arr = pos_vec_bi_arr[2][:]


	def compute_E(self, M, rtol=1e-6):
		root = optimize.newton(lambda E: E - self.e*np.sin(E) - M, x0=M)
		return root

	def compute_theta_star(self, E):
		return 2*np.arctan(np.sqrt((1 + self.e) / (1 - self.e)) * np.tan(E/2))


class Test_compute_timed_orbit:
	approach = Approach("TITAN", v_inf_vec_icrf_kms=np.array([-0.910, 5.081, 4.710]), rp=(2575 + 1500) * 1e3,
						psi=3 * np.pi / 2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=15000)
	orbiter.compute_timed_orbit_trajectory(t_seconds=75000, num_points=101)

	def test_timed_orbit(self):
		assert max(self.orbiter.t_array) == 75000







# SOURCE FILENAME : visibility.py
# AUTHOR          : Athul Pradeepkumar Girija, apradee@purdue.edu
# DATE CREATED    : 11/09/2022, 07:20 MT
# DATE MODIFIED   : 11/11/2022 15:16 MT
# REMARKS         : Compute the visibility of a target planet from a planetary lander,
#  					orbiter from a lander, and a target planet from an orbiter.

from astropy.coordinates import solar_system_ephemeris
from astropy.coordinates import get_body_barycentric_posvel
import numpy as np
from astropy.time import Time
from AMAT.planet import Planet


class Visibility:
	"""
	Base class used by other classes in AMAT.visibility.
	Contains functions to convert a vector from ICRF to BI frame.
	"""

	def ICRF_to_BI(self, X_ICRF):
		"""
		Converts an input vector from ICRF to body-inertial frame

		Parameters
		----------
		X_ICRF : numpy.ndarray
			row vector in ICRF frame

		Returns
		-------
		ans : numpy.ndarray
			input vector in body-inertial frame

		"""
		R1R3 = np.matmul(self.R1(np.pi/2 - self.d0), self.R3(np.pi/2 + self.a0))
		return np.matmul(R1R3, X_ICRF.T).T

	def R1(self, theta):
		"""
		Direction cosine matrix for rotation about x-axis.

		Parameters
		----------
		theta : float
			rotation angle about x-axis

		Returns
		-------
		ans : numpy.ndarray
			Direction cosine matrix for rotation about x-axis.

		"""
		return np.array([[1, 0, 0],
						 [0, np.cos(theta), np.sin(theta)],
						 [0, -np.sin(theta), np.cos(theta)]])

	def R2(self, theta):
		"""
		Direction cosine matrix for rotation about y-axis.

		Parameters
		----------
		theta : float
			rotation angle about y-axis

		Returns
		-------
		ans : numpy.ndarray
			Direction cosine matrix for rotation about y-axis.

		"""
		return np.array([[np.cos(theta), 0, -np.sin(theta)],
						 [0, 1, 0],
						 [np.sin(theta), 0, np.cos(theta)]])

	def R3(self, theta):
		"""
		Direction cosine matrix for rotation about z-axis.

		Parameters
		----------
		theta : float
			rotation angle about zaxis

		Returns
		-------
		ans : numpy.ndarray
			Direction cosine matrix for rotation about z-axis.

		"""
		return np.array([[np.cos(theta), np.sin(theta), 0],
						 [-np.sin(theta), np.cos(theta), 0],
						 [0, 0, 1]])


class LanderToPlanet(Visibility):
	"""
	Computes the range and elevation of a target planet from a site on an observer planet over
	one full rotation of the observer planet. This can be used to analyze the DTE data link from
	a stationary lander (say on Mars or Titan) to Earth.

	Attributes
	-----------
	planetObj :  AMAT.planet.Planet
		object for the observer planet
	observer_planet : str
		observer planet
	target_planet : str
		target planet
	a0 : float
		observer planet north pole right ascension wrt ICRF
	d0 : float
		observer planet north pole declination wrt ICRF
	arrival_date : astropy.time.Time object
		datetime object with scale=tdb
	longitude_array : numpy.ndarray
		body inertial longitude array, degrees
	t_array : numpy.ndarray
		array containing timestamps at which lander position is computed, seconds
	range : float
		range from observer to target, meters
	beta_array : numpy.ndarray
		beta angle array, degrees
	elevation_array : numpy.ndarray
		elevation angle array, degrees
	visible_array : numpy.ndarray
		boolean array containing visibility of lander to orbiter, True or False
	"""

	def __init__(self, target_planet, observer_planet, latitude, date,
				 num_points=101, ephem_file='spice-data/de432s.bsp'):
		"""
		Parameters
		----------
		target_planet : str
			Name of the planetary body whose elevation in the sky will be computed.
			Must be all uppercase. Valid entries are: 'VENUS', 'EARTH', 'MARS',
			'JUPITER', 'SATURN', 'TITAN', 'URANUS', 'NEPTUNE'
		observer_planet : str
			Name of the planetary body on which the lander/observer is located.
			Must be all uppercase. Valid entries are: 'VENUS', 'EARTH', 'MARS',
			'JUPITER', 'SATURN', 'TITAN', 'URANUS', 'NEPTUNE'.
			Must be different from target_planet
		latitude : float
			latitude of the landing site on observer planet, degrees
		date : str
			Datetime in the format "2036-03-28 00:00:00"
		num_points : int
			number of points to compute over one full rotation, defaults to 101
		ephem_file : str
			ephemeris filepath to spice-data/de432.bsp
		"""

		self.planetObj = Planet(observer_planet)
		self.observer_planet = observer_planet
		self.target_planet = target_planet

		self.a0 = self.planetObj.a0
		self.d0 = self.planetObj.d0

		# Compute the Solar System barycentric position of the observer planet
		# Titan is a special case, set observer_planet to Saturn if lander is on Titan.
		if observer_planet == 'TITAN':
			self.observer_planet = 'SATURN'
		else:
			self.observer_planet = observer_planet

		self.arrival_date = Time(date, scale="tdb")

		# Compute the ephemeris for the observer and target planet
		with solar_system_ephemeris.set(ephem_file):
			self.observer_planet_eph = get_body_barycentric_posvel(self.observer_planet, self.arrival_date)
			self.target_planet_eph = get_body_barycentric_posvel(target_planet, self.arrival_date)

		# Compute observer and target Solar System barycentric position(s) in ICRF (km)
		self.obs_planet_pos = np.array([self.observer_planet_eph [0].x.value,
										self.observer_planet_eph [0].y.value,
										self.observer_planet_eph [0].z.value])
		self.tar_planet_pos = np.array([self.target_planet_eph[0].x.value,
										self.target_planet_eph[0].y.value,
										self.target_planet_eph[0].z.value])

		# Compute observer to target position vector in ICRF
		self.obs_tar_pos_vec_icrf = self.tar_planet_pos - self.obs_planet_pos
		self.obs_tar_pos_vec_icrf_unit = self.obs_tar_pos_vec_icrf / np.linalg.norm(self.obs_tar_pos_vec_icrf)

		# compute the range from observer to target, in meters
		self.range = np.linalg.norm(self.obs_tar_pos_vec_icrf)*1e3

		# Convert observer to target unit vector in ICRF to observer planet BI frame
		self.obs_tar_pos_vec_bi_unit = self.ICRF_to_BI(self.obs_tar_pos_vec_icrf_unit)

		self.longitude_array = np.linspace(0, 360, num_points)  # body inertial longitude array
		self.t_array = np.linspace(0, 2*np.pi/(self.planetObj.OMEGA), num_points)  # time array
		self.beta_array = np.array([])  # beta angle array
		self.elevation_array = np.array([])  # elevation angle array
		self.visible_array = np.array([], dtype=bool)

		# Compute the elevation angle of the target for each longitude in longitude_array
		for longitude in self.longitude_array:
			self.lander_pos_x_bi = self.planetObj.RP*np.cos(latitude*np.pi/180)*np.cos(longitude*np.pi/180)
			self.lander_pos_y_bi = self.planetObj.RP*np.cos(latitude*np.pi/180)*np.sin(longitude*np.pi/180)
			self.lander_pos_z_bi = self.planetObj.RP*np.sin(latitude*np.pi/180)

			# Compute the lander zenith direction unit vector in the BI frame
			self.lander_pos_vec_bi = np.array([self.lander_pos_x_bi, self.lander_pos_y_bi, self.lander_pos_z_bi])
			self.lander_pos_vec_bi_unit = self.lander_pos_vec_bi / np.linalg.norm(self.lander_pos_vec_bi)

			# compute angle beta between lander zenith and observer to target unit vector
			beta = np.arccos(np.dot(self.lander_pos_vec_bi_unit, self.obs_tar_pos_vec_bi_unit))*180/np.pi
			elev = max(0, 90 - beta)  # if negative, target is below the horizon: set to 0

			if elev > 0:
				visible = True
			else:
				visible = False

			self.beta_array = np.append(self.beta_array, beta)
			self.elevation_array = np.append(self.elevation_array, elev)
			self.visible_array = np.append(self.visible_array, visible)


class LanderToOrbiter:
	"""
	Computes the range and elevation of an orbiter from a landing site on an observer planet over
	for a specified duration of time in seconds. This can be used to analyze the lander to relay orbiter
	data link from a stationary lander (say on Mars or Titan).

	Attributes
	------------
	t_array : numpy.ndarray
		array containing timestamps at which lander and orbiter positions are computed, seconds
	longitude_rate : float
		longitude advance rate, deg/s
	longitude_array : numpy.ndarray
		array containing longitudes at which positions are computed, degrees
	lander_pos_x_bi_arr : numpy.ndarray
		array containing lander position X coordinate in body inertial frame, meters
	lander_pos_y_bi_arr : numpy.ndarray
		array containing lander position Y coordinate in body inertial frame, meters
	lander_pos_z_bi_arr : numpy.ndarray
		array containing lander position Z coordinate in body inertial frame, meters
	orbiter_pos_x_bi_arr : numpy.ndarray
		array containing orbiter position X coordinate in body inertial frame, meters
	orbiter_pos_y_bi_arr : numpy.ndarray
		array containing orbiter position Y coordinate in body inertial frame, meters
	orbiter_pos_z_bi_arr : numpy.ndarray
		array containing orbiter position Z coordinate in body inertial frame, meters
	range_array : numpy.ndarray
		array containing lander-to-orbiter range values, meters
	beta_array : numpy.ndarray
		array containing lander-to-orbiter beta angle values, degrees
	elevation_array : numpy.ndarray
		array containing lander-to-orbiter elevation angle values, degrees
	visible_array : numpy.ndarray
		boolean array containing visibility of lander to orbiter, True or False
	"""

	def __init__(self, planet, latitude, orbiter, t_seconds, num_points=1600):
		"""

		Parameters
		----------
		planet : str
			Name of the planetary body on which the lander is located.
			Must be one of 'VENUS', 'EARTH', 'MARS', 'JUPITER', 'SATURN',
			'TITAN', 'URANUS', 'NEPTUNE'
		latitude : float
			latitude of the landing site on the planet, degrees
		orbiter : AMAT.orbiter.PropulsiveOrbiter
			propulsive orbiter object initialized using an approach trajectory
		t_seconds : float
			time in seconds for which elevation and range are to be computed
		num_points : int
			number of grid points to use in the time interval
		"""

		self.planetObj = Planet(planet)
		self.orbiterObj = orbiter
		# Compute the BI coordinates of the orbiter as a function of time.
		self.orbiterObj.compute_timed_orbit_trajectory(t_seconds, num_points=num_points)

		# First compute the BI coordinates of the lander as a function of time.
		# longitude advance rate = 360 / (2pi/OMEGA) deg/s
		self.t_array = np.linspace(0, t_seconds, num_points)
		self.longitude_rate = 360 / ((2*np.pi)/self.planetObj.OMEGA)
		self.max_longitude = self.longitude_rate * t_seconds
		self.longitude_array = np.linspace(0, self.max_longitude, num_points)  # body inertial longitude array

		self.lander_pos_x_bi_arr = np.array([])
		self.lander_pos_y_bi_arr = np.array([])
		self.lander_pos_z_bi_arr = np.array([])

		self.range_array = np.array([])  # range array
		self.beta_array = np.array([])  # beta angle array
		self.elevation_array = np.array([])  # elevation angle array
		self.visible_array = np.array([], dtype=bool)

		for longitude in self.longitude_array:
			self.lander_pos_x_bi = self.planetObj.RP*np.cos(latitude*np.pi/180)*np.cos(longitude*np.pi/180)
			self.lander_pos_y_bi = self.planetObj.RP*np.cos(latitude*np.pi/180)*np.sin(longitude*np.pi/180)
			self.lander_pos_z_bi = self.planetObj.RP*np.sin(latitude*np.pi/180)

			self.lander_pos_x_bi_arr = np.append(self.lander_pos_x_bi_arr, self.lander_pos_x_bi)
			self.lander_pos_y_bi_arr = np.append(self.lander_pos_y_bi_arr, self.lander_pos_y_bi)
			self.lander_pos_z_bi_arr = np.append(self.lander_pos_z_bi_arr, self.lander_pos_z_bi)

		self.orbiter_pos_x_bi_arr = self.planetObj.RP * self.orbiterObj.x_timed_orbit_arr
		self.orbiter_pos_y_bi_arr = self.planetObj.RP * self.orbiterObj.y_timed_orbit_arr
		self.orbiter_pos_z_bi_arr = self.planetObj.RP * self.orbiterObj.z_timed_orbit_arr

		# First, compute the lander zenith direction unit vector in the BI frame
		# Second, compute the orbiter direction unit vector in the BI frame
		# Third compute the lander to orbiter direction unit vector in the BI fra,e
		# Third, do a dot product of the vectors to compute beta angle
		# Elevation angle = 90 - beta
		for i in np.arange(0, num_points):
			lander_vec = np.array([self.lander_pos_x_bi_arr[i], self.lander_pos_y_bi_arr[i], self.lander_pos_z_bi_arr[i]])
			lander_unit = lander_vec / np.linalg.norm(lander_vec)

			orb_vec = np.array([self.orbiter_pos_x_bi_arr[i], self.orbiter_pos_y_bi_arr[i], self.orbiter_pos_z_bi_arr[i]])
			orb_unit = orb_vec / np.linalg.norm(orb_vec)

			lander_orb_vec = orb_vec - lander_vec
			lander_orb_vec_unit = lander_orb_vec / np.linalg.norm(lander_orb_vec)

			range = np.linalg.norm(lander_orb_vec)
			beta = np.arccos(np.dot(lander_unit, lander_orb_vec_unit))*180/np.pi
			elev = max(0, 90 - beta)
			if elev > 0:
				visible = True
			else:
				visible = False

			self.range_array = np.append(self.range_array, range)
			self.beta_array = np.append(self.beta_array, beta)
			self.elevation_array = np.append(self.elevation_array, elev)
			self.visible_array = np.append(self.visible_array, visible)


class OrbiterToPlanet(Visibility):
	"""
	Computes the visibility of a target planet from an orbiter at another planet for a specified
	duration of time in seconds. The target planet will be periodically eclipsed by the planet around
	which the spacecraft is orbiting. This can be used to analyze the relay orbiter data link to Earth.

	Attributes
	------------
	range : float
		range from observer to target, meters
	obs_tar_pos_vec_bi_unit : numpy.ndarray
		observer to target unit vector in BI frame
	t_array : numpy.ndarray
		array containing timestamps at which target planet visibility is computed, seconds
	visible_array : numpy.ndarray
		boolean array containing visibility of orbiter to target planet, True or False
	"""

	def __init__(self, target_planet, observer_planet, orbiter, date, t_seconds,
				 num_points=1600, ephem_file='spice-data/de432s.bsp'):
		"""

		Parameters
		----------
		target_planet : str
			Name of the target planetary body whose visibility is to be computed from the observer planet.
			Must be one of 'VENUS', 'EARTH', 'MARS', 'JUPITER', 'SATURN',
			'TITAN', 'URANUS', 'NEPTUNE'
		observer_planet : str
			Name of the observer planetary body at which the orbiter is located.
			Must be one of 'VENUS', 'EARTH', 'MARS', 'JUPITER', 'SATURN',
			'TITAN', 'URANUS', 'NEPTUNE'. Must be different from target_planet
		orbiter : AMAT.orbiter.PropulsiveOrbiter
			propulsive orbiter object initialized using an approach trajectory
		date : str
			Datetime in the format "2036-03-28 00:00:00"
		t_seconds : float
			time in seconds for which elevation and range are to be computed
		num_points : int
			number of grid points to use in the time interval
		ephem_file : str
			ephemeris filepath to spice-data/de432.bsp
		"""

		self.planetObj = Planet(observer_planet)
		self.observer_planet = observer_planet
		self.target_planet = target_planet

		self.orbiterObj = orbiter
		# Compute the BI coordinates of the orbiter as a function of time.
		self.orbiterObj.compute_timed_orbit_trajectory(t_seconds, num_points=num_points)

		self.a0 = self.planetObj.a0
		self.d0 = self.planetObj.d0

		self.t_array = np.linspace(0, t_seconds, num_points)

		# Compute the Solar System barycentric position of the observer planet
		# Titan is a special case, set observer_planet to Saturn if lander is on Titan.
		if observer_planet == 'TITAN':
			self.observer_planet = 'SATURN'
		else:
			self.observer_planet = observer_planet

		self.arrival_date = Time(date, scale="tdb")

		# Compute the ephemeris for the observer and target planet
		with solar_system_ephemeris.set(ephem_file):
			self.observer_planet_eph = get_body_barycentric_posvel(self.observer_planet, self.arrival_date)
			self.target_planet_eph = get_body_barycentric_posvel(target_planet, self.arrival_date)

		# Compute observer and target Solar System barycentric position(s) in ICRF (km)
		self.obs_planet_pos = np.array([self.observer_planet_eph[0].x.value,
										self.observer_planet_eph[0].y.value,
										self.observer_planet_eph[0].z.value])
		self.tar_planet_pos = np.array([self.target_planet_eph[0].x.value,
										self.target_planet_eph[0].y.value,
										self.target_planet_eph[0].z.value])

		# Compute observer to target position vector in ICRF
		self.obs_tar_pos_vec_icrf = self.tar_planet_pos - self.obs_planet_pos
		self.obs_tar_pos_vec_icrf_unit = self.obs_tar_pos_vec_icrf / np.linalg.norm(self.obs_tar_pos_vec_icrf)

		# compute the range from observer to target, in meters
		self.range = np.linalg.norm(self.obs_tar_pos_vec_icrf) * 1e3

		# Convert observer to target unit vector in ICRF to observer planet BI frame
		self.obs_tar_pos_vec_bi_unit = self.ICRF_to_BI(self.obs_tar_pos_vec_icrf_unit)

		# Compute the position of the orbiter in the BI frame as a function of time
		self.orbiter_pos_x_bi_arr = self.orbiterObj.x_timed_orbit_arr
		self.orbiter_pos_y_bi_arr = self.orbiterObj.y_timed_orbit_arr
		self.orbiter_pos_z_bi_arr = self.orbiterObj.z_timed_orbit_arr

		self.d_array = np.array([])
		self.visible_array = np.array([], dtype=bool)

		# Compute the distance between the origin at the line through the orbiter position in the direction of target
		for i in np.arange(0, num_points):
			a_vec = np.array([self.orbiter_pos_x_bi_arr[i],
							  self.orbiter_pos_y_bi_arr[i],
							  self.orbiter_pos_z_bi_arr[i]])
			p_vec = np.array([0, 0, 0])
			n_vec = self.obs_tar_pos_vec_bi_unit

			d = np.linalg.norm(np.cross(p_vec - a_vec, n_vec))
			if d < 1.0 and np.dot(self.obs_tar_pos_vec_bi_unit, -1*a_vec) > 0:
				visible = False
			else:
				visible = True

			self.d_array = np.append(self.d_array, d)
			self.visible_array = np.append(self.visible_array, visible)



import numpy as np
from AMAT.approach import Approach
from AMAT.orbiter import PropulsiveOrbiter

try:
	from AMAT.visibility import LanderToPlanet, LanderToOrbiter, OrbiterToPlanet
except ModuleNotFoundError:
	raise ModuleNotFoundError("Cannot import LanderToPlanet, LanderToOrbiter, OrbiterToPlanet from AMAT.visibility")


class Test_Titan_Earth_Equatorial:
	visibility = LanderToPlanet(observer_planet="TITAN", target_planet="EARTH", latitude=3.0, date="2034-03-28 00:00:00")

	def test_range(self):
		assert abs(self.visibility.range/1.5e11 - 8.836) < 0.01

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 61.87) < 0.1


class Test_Titan_Earth_High_Latitude_North:
	visibility = LanderToPlanet(observer_planet="TITAN", target_planet="EARTH", latitude=60.0, date="2034-03-28 00:00:00")

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 4.87) < 0.1


class Test_Earth_Mars_40_North:
	visibility = LanderToPlanet(observer_planet="EARTH", target_planet="MARS", latitude=40, date="2022-11-09 00:00:00")

	def test_range(self):
		assert abs(self.visibility.range/1.4959e11 - 0.587) < 0.01

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 74.18) < 0.1


class Test_Earth_Jupiter_40_North:
	visibility = LanderToPlanet(observer_planet="EARTH", target_planet="JUPITER", latitude=40, date="2022-11-09 00:00:00")

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 48.14) < 0.1


class Test_Earth_Saturn_40_North:
	visibility = LanderToPlanet(observer_planet="EARTH", target_planet="SATURN", latitude=40, date="2022-11-09 00:00:00")

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 33.49) < 0.1


class Test_Earth_Uranus_40_North:
	visibility = LanderToPlanet(observer_planet="EARTH", target_planet="URANUS", latitude=40, date="2022-11-09 00:00:00")

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 66.42) < 0.1

class Test_Earth_Neptune_40_North:
	visibility = LanderToPlanet(observer_planet="EARTH", target_planet="NEPTUNE", latitude=40, date="2022-11-09 00:00:00")

	def test_max_elevation(self):
		assert abs(max(self.visibility.elevation_array) - 45.90) < 0.1


class Test_OrbiterElevationRange_Titan:
	approach = Approach("TITAN", v_inf_vec_icrf_kms=np.array([-0.910, 5.081, 4.710]), rp=(2575+1500)*1e3, psi=3*np.pi/2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=15000)
	visibility = LanderToOrbiter(planet="TITAN", latitude=3.00, orbiter=orbiter, t_seconds=86400 * 16, num_points=2001)

	def test_elevation_and_range(self):
		assert abs(min(self.visibility.range_array) / 1e3 - 1520.05) < 0.1
		assert abs(max(self.visibility.elevation_array) - 86.69) < 0.1


class Test_OrbiterElevationRange_EQUATORIAL_ORBIT_EARTH:
	approach = Approach("EARTH", v_inf_vec_icrf_kms=np.array([1, 1, 0]), rp=(6371 + 250) * 1e3, psi=3*np.pi/2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=250)
	visibility = LanderToOrbiter(planet="EARTH", latitude=0.00, orbiter=orbiter, t_seconds=1200, num_points=1001)

	def test_elevation_and_range(self):
		assert abs(min(self.visibility.range_array)/1e3 - 250) < 0.1
		assert abs(max(self.visibility.elevation_array) - 90) < 1.0


class Test_OrbiterToPlanet:
	approach = Approach("TITAN", v_inf_vec_icrf_kms=np.array([-0.910, 5.081, 4.710]), rp=(2575+1500)*1e3, psi=3*np.pi/2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=15000)
	visibility = OrbiterToPlanet(target_planet="EARTH", observer_planet="TITAN",
								 orbiter=orbiter, date="2034-03-28 00:00:00", t_seconds=86400, num_points=1600)

	def test_visibility(self):
		assert True in self.visibility.visible_array
		assert False in self.visibility.visible_array



from AMAT.planet import Planet
from AMAT.vehicle import Vehicle

import numpy as np
from scipy import interpolate
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib import rcParams
from matplotlib.patches import Polygon

# Create a planet object
planet=Planet("VENUS")
planet.h_skip = 150000.0
planet.h_low=10.0E3

# Load an nominal atmospheric profile with height, temp, pressure, density data
planet.loadAtmosphereModel('../../../atmdata/Venus/venus-gram-avg.dat', 0 , 1 ,2, 3)

# Set up a vehicle object
vehicle=Vehicle('SmallSat1', 37, 20, 0.0, 1.767, 0.0, 0.35, planet)
vehicle.setInitialState(150.0,-6.22,23.16,10.8191,89.9982,-5.30,0.0,0.0)
vehicle.setSolverParams(1E-6)
vehicle.setDragModulationVehicleParams(20, 7.5)

# Set up the drag modulation entry phase guidance parameters.
vehicle.setDragEntryPhaseParams(2.0, 20.0, 101, -200.0)

# Set the target orbit parameters.
vehicle.setTargetOrbitParams(200.0, 1800.0, 20.0)

# Define the path to atmospheric files to be used for the Monte Carlo simulations.
atmfiles = ['../../../atmdata/Venus/LAT20N.txt']

# Set up the Monte Carlo simulation for drag modulation.
# NPOS = 151, NMONTE = 200
# Target EFPA = -5.30 deg
# EFPA 1-sigma error = +/- 0.033 deg
# Nominal beta_1 = 20 kg/m2
# beta_1 1-sigma = 0.0
# guidance time step for entry = 1.0s (Freq. = 1 Hz)
# guidance time step after jettison = 1.0 s
# max. solver time step = 0.1 s
# max. time used by solver = 2400 s

vehicle.setupMonteCarloSimulationD(151, 200, atmfiles, 0 , 1, 2, 3, 4, True,
                                   -5.20,  0.0333, 20.0, 0.0,
                                    1.0, 1.0, 0.1, 2400.0)
# Run 1000 trajectories
vehicle.runMonteCarloD2(1000, '../../../data/acta-astronautica/smallsat-mission-concepts/venus/MCB-Venus')



from AMAT.telecom import Link, Schedule
from AMAT.approach import Approach
from AMAT.orbiter import PropulsiveOrbiter
from AMAT.visibility import LanderToPlanet, LanderToOrbiter, OrbiterToPlanet
import numpy as np


class Test_Cassini_Earth_Downlink:
	link = Link(freq=8.425, Pt=20, Gt_dBi=47.20, Gr_dBi=68.24, Ts=40.5,
				range_km=1.5E9, L_other_dB=0.55, rate_kbps=14.00, Eb_N0_req=0.31)

	def test_Eb_N0(self):
		assert round(self.link.EB_N0, 2) == 1.94


class Test_Huygens_Cassini_Relay:
	link = Link(freq=2.040, Pt=11.7, Gt_dBi=1.05, Gr_dBi=34.7, Ts=230,
				range_km=80E3, La_dB=0.05, L_other_dB=3.3, rate_kbps=8.00, Eb_N0_req=2.55)

	def test_Eb_N0(self):
		assert round(self.link.EB_N0, 2) == 10.14


class Test_OCEANUS_Uranus_Earth_Downlink:
	link = Link(freq=32, Pt=40, Gt_dBi=60.33, Gr_dBi=80.0, Ts=46.97,
				range_km=3.05E9, L_other_dB=0.60, rate_kbps=34, Eb_N0_req=0.31)

	def test_Eb_N0(self):
		assert round(self.link.EB_N0, 2) == 7.54


class Test_OCEANUS_Saturn_Probe_Relay:
	link = Link(freq=0.405, Pt=10.7, Gt_dBi=5.10, Gr_dBi=22.3, Ts=183.0,
				range_km=1.62E6, La_dB=2.0,  L_other_dB=2.75, rate_kbps=0.2, L_DSN_dB=0, Eb_N0_req=5.00)

	def test_Eb_N0(self):
		assert round(self.link.EB_N0, 2) == 5.93


class Test_OCEANUS_Uranus_Probe_Relay:
	link = Link(freq=0.405, Pt=10.7, Gt_dBi=5.89, Gr_dBi=22.3, Ts=104.0,
				range_km=1.55E5, L_other_dB=1.34, rate_kbps=0.2, Eb_N0_req=5.00)

	def test_Eb_N0(self):
		assert round(self.link.EB_N0, 2) == 31.62


class Test_Dragonfly_Downlink:
	link = Link(freq=8.425, Pt=30, Gt_dBi=30, Gr_dBi=80.0, Ts=40.5,
				range_km=1.5E9, L_other_dB=0.54, rate_kbps=2.00, Eb_N0_req=0.31)

	def test_link_budget(self):
		pass


class Test_Titan_Lander_To_Relay_Orbiter_X_Uplink:
	link = Link(freq=8.425, Pt=30, Gt_dBi=30, Gr_dBi=32.0, Ts=230,
				range_km=15E3, La_dB=0.05, L_other_dB=3.3, rate_kbps=7500.00, Eb_N0_req=2.55)

	def test_Eb_N0(self):
		self.link.print_link_budget()

class Test_Relay_Orbiter_Earth_Ka_Downlink:
	link = Link(freq=32.00, Pt=40, Dt=4.0, eta_t=0.60, Gr_dBi=79.83, Ts=48.0,
				range_km=1.5E9, L_other_dB=0.50, rate_kbps=500, Eb_N0_req=0.31)

	def test_Eb_N0(self):
		self.link.print_link_budget()

class Test_WorldView_4_Earth_Downlink:
	link = Link(freq=8.185, Pt=7.5, Gt_dBi=29.1, Gr_dBi=48.0, Ts=273.0,
				range_km=2718.0, L_other_dB=2.00, rate_kbps=400E3, Eb_N0_req=11.9)

	def test_Eb_N0(self):
		self.link.print_link_budget()


class Test_Schedule:
	schedule = Schedule(86400*16, 2001, [np.array([0, 86400]), np.array([86400*2, 86400*3])])

	def test_dummy(self):
		pass


class Test_Data_Volume_DTE:
	visibility1 = LanderToPlanet(observer_planet="TITAN", target_planet="EARTH",
								 latitude=3.0, date="2034-03-28 00:00:00")
	link = Link(freq=8.425, Pt=30, Gt_dBi=30, Gr_dBi=80.0, Ts=40.5,
				range_km=1.5E9, L_other_dB=0.54, rate_kbps=2.00, Eb_N0_req=0.31)
	schedule1 = Schedule(86400*16, 101, [np.array([4*86400, 5*86400]),
										  np.array([6*86400, 7*86400]),
										  np.array([8*86400, 9*86400]),
										  np.array([10*86400, 11*86400])])
	link.compute_data_volume(visibility1, schedule1)

	def test_data_volume(self):
		pass


class Test_Data_Volume_To_Relay:
	approach = Approach("TITAN", v_inf_vec_icrf_kms=np.array([-0.910, 5.081, 4.710]),
						rp=(2575+1500) * 1e3, psi=3*np.pi/2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=15000)
	visibility1 = LanderToOrbiter(planet="TITAN", latitude=3.00, orbiter=orbiter, t_seconds=86400*16, num_points=2001)


	link = Link(freq=8.425, Pt=30, Gt_dBi=30, Gr_dBi=32.0, Ts=230, range_km=15E3,
				La_dB=0.05, L_other_dB=3.3, rate_kbps=7500.00, Eb_N0_req=2.55)
	schedule1 = Schedule(86400*16, 2001, [np.array([0.04*86400, 0.66*86400]),
										  np.array([0.92*86400, 1.30*86400])])
	link.compute_data_volume(visibility1, schedule1)

	def test_data_volume(self):
		pass


class Test_Relay_to_Earth_Downlink:
	approach = Approach("TITAN", v_inf_vec_icrf_kms=np.array([-0.910, 5.081, 4.710]),
						rp=(2575 + 1500) * 1e3, psi=3 * np.pi / 2)
	orbiter = PropulsiveOrbiter(approach=approach, apoapsis_alt_km=15000)
	visibility1 = OrbiterToPlanet(target_planet="EARTH", observer_planet="TITAN",
								  orbiter=orbiter, date="2034-03-28 00:00:00", t_seconds=86400*2, num_points=2001)

	link = Link(freq=32.00, Pt=40, Dt=4.0, eta_t=0.60, Gr_dBi=79.83, Ts=48.0,
				range_km=1.5E9, L_other_dB=0.50, rate_kbps=500, Eb_N0_req=0.31)
	schedule1 = Schedule(86400*2, 2001, [np.array([1.30*86400, 86400*16])])
	link.compute_data_volume(visibility1, schedule1)

	def test_data_volume(self):
		pass






import numpy as np
from scipy.interpolate import interp1d
from matplotlib import rcParams
import matplotlib.pyplot as plt
from scipy.integrate import cumulative_trapezoid


class Planet:
	"""
	The Planet class is used to store planetary constants, 
	load atmospheric data from look-up tables, and define
	non-dimensional parameters used in the simulations.
	
	Attributes
	----------
	ID : str
		String identifier of planet object
	RP : float
		Mean equatorial radius of the target planet in meters
	OMEGA : float
		Mean angular velocity of rotation of the planet about 
		its axis of rotation in rad/s
	GM : float
		Standard gravitational parameter of the planet in m3/s2
	rho0 : float
		Reference atmospheric density at the surface of the target 
		planet in kg/m3
	CPCV : float
		Specific heat ratio CP/CV at the surface of the planet
	J2 : float
		zonal harmonic coefficient J2
	J3 : float
		zonal harmonic coefficient J3
	a0 : float
		right ascension of North Pole in ICRF, rad
	d0 : float
		declination of North Pole in ICRF, rad
	h_thres : float
		Atmospheric model cutoff altitude in meters, 
		density is set to 0, if altitude exceeds h_thres
	h_skip : float
		If vehicle altitude exceeds this value, trajectory is cut off
		and vehicle is assumed to skip off the atmosphere
	h_trap : float
		If vehicle altitude falls below this value, trajectory is cut off
		and vehicle is assumed to hit the surface
	h_low : float
		If terminal altitude is below this value vehicle is assumed to
		be trapped in the atmosphere. 
	Vref : float
		Reference velocity for non-dimensionalization of entry equations
	tau : float
		Reference timescale used to non-dimensionalize time, angular rates
	OMEGAbar : float
		Reference non-dimensional angular rate of planet's rotation
	EARTHG : float
		Reference value of acceleration due to Earth's gravity
	ATM : numpy.ndarray
		Array containing the data loaded from atmospheric lookup file
	ATM_height : numpy.ndarray
		Array containing height values from atm. look up dat file
	ATM_temp : numpy.ndarray
		Array containing temperature values from atm. look up dat file
	ATM_pressure : numpy.ndarray
		Array containing pressure values from atm. look up dat file
	ATM_density : numpy.ndarray
		Array containing density values from atm. look up dat file
	ATM_sonic : numpy.ndarray
		Array containing computed sonic speed values
	temp_int : scipy.interpolate.interpolate.interp1d
		Function which interpolates temperature as function of height
	pressure_int : scipy.interpolate.interpolate.interp1d
		Function which interpolates pressure as function of height
	density_int : scipy.interpolate.interpolate.interp1d
		Function which interpolates density as function of height
	sonic_int : scipy.interpolate.interpolate.interp1d
		Function which interpolates sonic speed as function of height
	"""

	def __init__(self, planetID):
		"""
		Initializes the planet object with the planetary constants.
		
		Parameters
		----------
		planetID : str
			Name of the planetary body, must be all uppercase; 
			Valid entries are: 'VENUS', 'EARTH', 'MARS',
			'JUPITER', 'SATURN', 'TITAN', 'URANUS', 'NEPTUNE'

		"""

		if planetID == 'VENUS':

			self.ID = 'VENUS'
			self.RP = 6051.8000E3
			self.OMEGA = -2.99237E-7
			self.GM = 3.248599E14
			self.rho0 = 64.790
			self.CPCV = 1.289
			self.J2 = 4.458E-6
			self.J3 = 0.000000
			self.a0 = 272.76 * np.pi / 180.0
			self.d0 = 67.16 * np.pi / 180.0
			self.h_thres = 180000.0
			self.h_skip = 180000.0     
			self.h_trap = 10000.0
			self.h_low = 60.0E3

		elif planetID == 'EARTH':

			self.ID = 'EARTH'
			self.RP = 6371.0000E3
			self.OMEGA = 7.272205E-5
			self.GM = 3.986004E14
			self.rho0 = 1.2250
			self.CPCV = 1.4
			self.J2 = 1082.6E-6
			self.J3 = -2.532E-6
			self.a0 = 0.0 * np.pi / 180.0
			self.d0 = 90.0 * np.pi / 180.0
			self.h_thres = 120.0E3
			self.h_skip = 120.0E3      
			self.h_trap = 10.0E3
			self.h_low = 50.0E3

		elif planetID == 'MARS':

			self.ID = 'MARS'
			self.RP = 3389.5000E3
			self.OMEGA = 7.088253E-5
			self.GM = 4.282837E13
			self.rho0 = 0.0200
			self.CPCV = 1.289
			self.J2 = 1960.45E-6
			self.J3 = 31.5E-6
			self.a0 = 317.68143 * np.pi / 180.0
			self.d0 = 52.88650 * np.pi / 180.0
			self.h_thres = 120.0E3
			self.h_skip = 120.0E3      
			self.h_trap = 10.0E3 
			self.h_low = 50.0E3

		elif planetID == 'JUPITER':

			self.ID = 'JUPITER'
			self.RP = 69911.0E3
			self.OMEGA = 1.758518E-04
			self.GM = 1.26686534E17
			self.rho0 = 0.16288
			self.CPCV = 1.4348
			self.J2 = 14736E-6
			self.J3 = 0.0
			self.a0 = 268.056595 * np.pi / 180.0
			self.d0 = 64.495303 * np.pi / 180.0
			self.h_thres = 1000.0E3
			self.h_skip = 1000.0E3     
			self.h_trap = 50.0E3 

		elif planetID == 'SATURN':
			self.ID = 'SATURN'
			self.RP = 58232.0E3
			self.OMEGA = 1.6379E-04
			self.GM = 3.7931187E16
			self.rho0 = 0.19847
			self.CPCV = 1.4348
			self.J2 = 16298E-6
			self.J3 = 0.0
			self.a0 = 40.589 * np.pi / 180.0
			self.d0 = 83.537 * np.pi / 180.0
			self.h_thres = 1000.0E3
			self.h_skip = 1000.0E3     
			self.h_trap = 50.0E3      

		elif planetID == 'TITAN':

			self.ID = 'TITAN'
			self.RP = 2575.0000E3
			self.OMEGA = 4.5451280E-6
			self.GM = 8.9780000E12
			self.rho0 = 5.43500
			self.CPCV = 1.400
			self.J2 = 31.808E-6
			self.J3 = -1.880E-6
			self.a0 = 39.4827 * np.pi / 180.0
			self.d0 = 83.4279 * np.pi / 180.0
			self.h_thres = 1000.0E3
			self.h_skip = 1000.0E3      
			self.h_trap = 30.0E3
			self.h_low = 300.0E3

		elif planetID == 'URANUS':
			self.ID = 'URANUS'
			self.RP = 25559.0E3
			self.OMEGA = -1.01237E-4
			self.GM = 5.793939E15
			self.rho0 = 0.3788
			self.CPCV = 1.450
			self.J2 = 3343.3E-6
			self.J3 = 0.000000
			self.a0 = 257.311 * np.pi / 180.0
			self.d0 = -15.175 * np.pi / 180.0
			self.h_thres = 1500.0E3
			self.h_skip = 1500.0E3      
			self.h_trap = 10.0E3
			self.h_low = 10.0E3

		elif planetID == 'NEPTUNE':
			self.ID = 'NEPTUNE'
			self.RP = 24622.000E3
			self.OMEGA = 1.083385E-4
			self.GM = 6.8365299E15
			self.rho0 = 0.44021
			self.CPCV = 1.450
			self.J2 = 3411.0E-6
			self.J3 = 0.000000
			self.a0 = 299.36 * np.pi / 180.0
			self.d0 = 43.36 * np.pi / 180.0
			self.h_thres = 1000.0E3
			self.h_skip = 1000.0E3      
			self.h_trap = 10.0E3
			self.h_low = 100.0E3

		else:
			raise ValueError(" >>> ERR : Invalid planet identifier provided.\
							Valid entries are: VENUS, EARTH, MARS, \
							JUPITER, SATURN, TITAN, URANUS, NEPTUNE")

		# compute reference values for planet
		if self.ID:
			self.Vref = np.sqrt(self.GM/self.RP)
			self.tau = self.RP / self.Vref
			self.OMEGAbar = self.OMEGA*self.tau
			self.EARTHG = 9.80665

		self.ATM = None
		self.ATM_height = None
		self.ATM_temp = None
		self.ATM_pressure = None
		self.ATM_density = None
		self.ATM_sonic = None
		self.temp_int = None
		self.pressure_int = None
		self.density_int = None
		self.sonic_int = None

	def loadAtmosphereModel(self, datfile, heightCol, tempCol, presCol, densCol, intType='cubic', heightInKmFlag=False):
		"""
		Load atmospheric model from a look up table with 
		height, temperature, pressure, and density
		
		Parameters
		----------
		datfile : str
			file containing atmospheric lookup table
		heightCol : int
			column number of height values, assumes unit = meters 
			(first column = 0, second column = 1, etc.)
		tempCol : int
			column number of temperature values, assumes unit = K
			(first column = 0, second column = 1, etc.)
		presCol : int
			column number of pressure values, assumes unit = Pascals 
			(first column = 0, second column = 1, etc.)
		densCol : int
			column number of density values, assumes unit = kg/m3 
			(first column = 0, second column = 1, etc.)
		intType : str, optional
			interpolation type: 'linear', 'quadratic' or 'cubic'
			defaults to 'cubic'
		heightInKmFlag : bool, optional
			optional, set this to True if heightCol has units of km, 
			False by default
		"""
	
		self.ATM = np.loadtxt(datfile)

		if heightInKmFlag is True:
			# convert heightCol from km to meters
			self.ATM_height = self.ATM[:, heightCol]*1E3
		else:
			self.ATM_height = self.ATM[:, heightCol]

		self.ATM_temp = self.ATM[:, tempCol]
		self.ATM_pressure = self.ATM[:, presCol]
		self.ATM_density = self.ATM[:, densCol]

		# derive speed of sound profile from pressure, density 
		# and specific heat ratio
		self.ATM_sonic = np.sqrt(self.CPCV*self.ATM_pressure/self.ATM_density)
		
		# create interpolating functions using scipy.interpolate.interp1d
		# fill_value is the value returned by the interpolating function if 
		# input arguments fall outside available data range.
		# fill_value = 0.0 for temp_int, pressure_int, density_int

		# bounds_error=False indicates the function will not return an error 
		# if input arguments fall outside available data range.

		# The fill_value and bounds_error arguments are used due to the fact that 
		# while propogating trajectories, the vehicle
		# might go above the altitude for which atmospheric data is available, or 
		# go below the surface where at atmpospheric data is available.

		self.temp_int = interp1d(self.ATM_height, self.ATM_temp, kind=intType, fill_value=0.0, bounds_error=False)
		self.pressure_int = interp1d(self.ATM_height, self.ATM_pressure, kind=intType, fill_value=0.0, bounds_error=False)
		self.density_int = interp1d(self.ATM_height, self.ATM_density, kind=intType, fill_value=0.0, bounds_error=False)
		self.sonic_int = interp1d(self.ATM_height, self.ATM_sonic, kind=intType, fill_value=1E20, bounds_error=False)

	def density(self, h):
		"""
		Returns atmospheric density, scalar value, 
		at altitude h (in meters)

		
		Parameters
		----------
		h : float
			altitude in meters

		Returns
		----------
		ans : float
			atmospheric density at height h
		"""

		if 0 <= h <= self.h_thres:
			# if altitude is within available data range, return
			# density data from interpolating function density_int()
			return float(self.density_int(h))

		elif h > self.h_thres:
			# if altitude is above atmospheric 
			# cut off altitude, return density=0
			return 0

		elif h < 0:
			# if altitude is below 0, return the reference 
			# density at the surface the trajectory will be cut off at
			# the surface / trap in altitude during post-processing 
			# rho0 is provided so that the solver has some numerical 
			# density value to work with even if the
			# trajectory goes below the surface during propagation
			return self.rho0

	def tempvectorized(self, h):
		"""
		Returns atmospheric temperature, vector
		at altitudes array h[:] in meters

		
		Parameters
		----------
		h : numpy.ndarray
			altitude h[:] at which atmospheric temperature is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric temperature at altitudes h[:], K
		"""		
		ans = np.zeros(len(h))
		ans[:] = self.temp_int(h[:])
		
		return ans

	def presvectorized(self, h):
		"""
		Returns atmospheric pressure, vector
		at altitudes array h[:] in meters

		Parameters
		----------
		h : numpy.ndarray
			altitude h[:] at which atmospheric pressure is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric pressure at altitudes h[:], K
		"""		
		ans = np.zeros(len(h))
		ans[:] = self.pressure_int(h[:])
		
		return ans

	def densityvectorized(self, h):
		"""
		Returns atmospheric density, vector
		at altitudes array h[:] in meters

		Parameters
		----------
		h : numpy.ndarray
			altitude h[:] at which atmospheric density is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric density at altitudes h[:], K
		"""		

		ans = np.zeros(len(h))
		ans[:] = self.density_int(h[:])
		
		return ans

	def avectorized(self, h):
		"""
		Returns atmospheric sonic speed, vector
		at altitudes array h[:] in meters

		Parameters
		----------
		h : numpy.ndarray
			altitude h[:] at which sonic speed is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the sonic speed at altitudes h[:], K
		"""		

		ans = np.zeros(len(h))
		ans[:] = self.sonic_int(h[:])
		
		return ans

	def rho(self, r, theta, phi):
		"""
		Returns atmospheric density rho, scalar, as a function
		of radial distance from the target planet center r
		as well as longitude theta and latitude phi
		
		
		Parameters
		----------
		r : float
			radial distance r measured from the planet center
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		
		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric density at (r,theta,phi)
		"""		
		h = r - self.RP        # compute altitude from radial distance
		ans = self.density(h)  # compute density
		
		return ans

	def rhovectorized(self, r):
		"""
		Returns atmospheric density, vector
		at radial distance array r[:] in meters

		Parameters
		----------
		r : numpy.ndarray
			radial distance r[:] at which density is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric density at radial distance r[:]
		"""		
		h = np.zeros(len(r))
		ans = np.zeros(len(r))
		RP_vec = np.ones(len(r))*self.RP
		h[:] = r[:] - RP_vec[:]
		ans[:] = self.density_int(h[:])

		return ans

	def pressurevectorized(self, r):
		"""
		Returns atmospheric pressure, vector
		at radial distance array r[:] in meters

		Parameters
		----------
		r : numpy.ndarray
			radial distance r[:] at which pressure is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric pressure at radial distance r[:]
		"""		
		h = np.zeros(len(r))
		ans = np.zeros(len(r))
		RP_vec = np.ones(len(r))*self.RP
		h[:] = r[:] - RP_vec[:]
		ans[:] = self.pressure_int(h[:])
		return ans

	def temperaturevectorized(self, r):
		"""
		Returns atmospheric temperature, vector
		at radial distance array r[:] in meters

		Parameters
		----------
		r : numpy.ndarray
			radial distance r[:] at which temperature is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric temperature at radial distance r[:]
		"""		
		h = np.zeros(len(r))
		ans = np.zeros(len(r))
		RP_vec = np.ones(len(r))*self.RP
		h[:] = r[:] - RP_vec[:]
		ans[:] = self.temp_int(h[:])
		return ans

	def sonicvectorized(self, r):
		"""
		Returns atmospheric sonic speed, vector
		at radial distance array r[:] in meters

		Parameters
		----------
		r : numpy.ndarray
			radial distance r[:] at which sonic speed is desired

		Returns
		----------
		ans : numpy.ndarray
			returns the atmospheric speed at radial distance r[:]
		"""
		h = np.zeros(len(r))
		ans = np.zeros(len(r))
		RP_vec = np.ones(len(r))*self.RP
		h[:] = r[:] - RP_vec[:]
		ans[:] = self.sonic_int(h[:])
		return ans

	def rbar(self, r):
		"""
		Returns non-dimensional rbar=r/RP
		
		Parameters
		----------
		r : float
			radial distance in meters

		Returns
		----------
		ans : float
			non-dimensional rbar
		"""
		ans = r/self.RP
		return ans

	def rho2(self, rbar, theta, phi):
		"""
		Returns atmospheric density rho, scalar, as a function
		of non-dimensional radial distance rbar, longitude theta, 
		and latitude phi
		
		Parameters
		----------
		rbar : float
			nondimensional radial distance rbar
			measured from the planet center
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		
		Returns
		----------
		ans : float
			returns the atmospheric density at (rbar,theta,phi)
		"""		
		r = rbar*self.RP
		ans = self.rho(r, theta, phi)
		return ans

	def rhobar(self, rbar, theta, phi):
		"""
		Returns non-dimensional density rhobar = rho / rho0
		as a function of non-dimensional radial distance rbar, 
		longitude theta, and latitude phi
		
		Parameters
		----------
		rbar : float
			nondimensional radial distance rbar
			measured from the planet center
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		
		Returns
		----------
		ans : float
			non-dimensional density at (rbar,theta,phi)
		"""		
		ans = self.rho2(rbar, theta, phi)/self.rho0
		return ans

	def checkAtmProfiles(self, h0=0.0, dh=1000.0):
		"""
		Function to check the loaded atmospheric profile data.
		Plots temperature, pressure, density and sonic speed
		as function of altitude.

		Parameters
		----------
		h0 : float, optional
			lower limit of altitude, defaults to 0.0
		dh : float, optional
			height interval
		
		Returns
		----------
		A plot showing the atmospheric profiles loaded
		from the lookup tables
		"""
		h_array = np.linspace(h0, self.h_thres, int(self.h_thres/dh))

		# compute profiles of T, P, rho, a using interpolated functions
		# vectorized functions are more efficient for this kind of computation.
		
		T_array = self.tempvectorized(h_array)
		P_array = self.presvectorized(h_array)
		r_array = self.densityvectorized(h_array)
		a_array = self.avectorized(h_array)

		fig = plt.figure()
		fig.set_size_inches([6.5, 6.5])
		rcParams['font.family'] = 'sans-serif'
		rcParams['font.sans-serif'] = ['DejaVu Sans']

		plt.subplot(2, 2, 1)
		plt.plot(T_array, h_array*1E-3, 'r-', linewidth=2.0)
		plt.xlabel("Temperature, K", fontsize=12)
		plt.ylabel("Altitude, km", fontsize=12)
		plt.xticks(fontsize=12)
		plt.yticks(fontsize=12)
		plt.grid('on', linestyle='-', linewidth=0.2)
		
		plt.subplot(2, 2, 2)
		plt.plot(P_array*1E-3, h_array*1E-3, 'r-', linewidth=2.0)
		plt.xlabel("Pressure, kPa", fontsize=12)
		plt.ylabel("Altitude, km", fontsize=12)
		plt.xscale('log')
		plt.xticks(fontsize=12)
		plt.yticks(fontsize=12)
		plt.grid('on', linestyle='-', linewidth=0.2)

		plt.subplot(2, 2, 3)
		plt.plot(r_array, h_array*1E-3, 'r-', linewidth=2.0)
		plt.xlabel("Density, kg/m3", fontsize=12)
		plt.ylabel("Altitude, km", fontsize=12)
		plt.xscale('log')
		plt.xticks(fontsize=12)
		plt.yticks(fontsize=12)
		plt.grid('on', linestyle='-', linewidth=0.2)

		plt.subplot(2, 2, 4)
		plt.plot(a_array, h_array*1E-3, 'r-', linewidth=2.0)
		plt.xlabel("Speed of Sound, m/s", fontsize=12)
		plt.ylabel("Altitude, km", fontsize=12)
		plt.xticks(fontsize=12)
		plt.yticks(fontsize=12)
		plt.grid('on', linestyle='-', linewidth=0.2)

		ax = plt.gca()
		ax.tick_params(direction='in')
		ax.yaxis.set_ticks_position('both')
		ax.xaxis.set_ticks_position('both')

		plt.tight_layout()
		plt.show()

	def computeR(self, h):
		"""
		Returns radial distance r, as 
		a function of altitude h, METERS

		Parameters
		----------
		h : float
			altitude in meters

		Returns
		----------
		r : float
			radial distance r=RP+h
		"""
		r = self.RP + h
		return r

	def computeH(self, r):
		"""
		Returns altitude h, as 
		a function of radial distance r, METERS

		Parameters
		----------
		r : float
			radial distance in meters

		Returns
		----------
		h : float
			h = r - RP
		"""
		h = r - self.RP
		return h

	def nonDimState(self, r, theta, phi, v, psi, gamma, drange):
		"""
		Computes non-dimensional trajectory state variables from 
		dimensional trajectory state variables

		Parameters
		----------
		r : float
			radial distance in meters
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		v : float
			planet-relative speed, m/s
		psi : float
			heading angle, radians
		gamma : float
			planet-relative flight-path angle, radians
		drange : float
			downrange distance measured from entry-interface

		Returns
		----------
		rbar : float
			non-dimensional radial distance in meters
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		vbar : float
			non-dimensional planet-relative speed, m/s
		psi : float
			heading angle, radians
		gamma : float
			planet-relative flight-path angle, radians
		drangebar : float
			non-dimensional downrange distance measured from 
			entry-interface
		"""
		rbar = r / self.RP           # Non-dimensional entry radius
		vbar = v / self.Vref         # Non-dimensional entry velocity
		drangebar = drange / self.RP      # Non-dimensional entry downrange

		return rbar, theta, phi, vbar, psi, gamma, drangebar

	def dimensionalize(self, tbar, rbar, theta, phi, vbar, psi, gamma, drangebar):
		"""
		Computes dimensional trajectory state variables from 
		non-dimensional trajectory state variables

		Parameters
		----------
		rbar : float
			non-dimensional radial distance in meters
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		vbar : float
			non-dimensional planet-relative speed, m/s
		psi : float
			heading angle, radians
		gamma : float
			planet-relative flight-path angle, radians
		drangebar : float
			non-dimensional downrange distance measured from 
			entry-interface

		Returns
		----------
		r : float
			radial distance in meters
		theta : float
			longitude theta(RADIANS), theta in [-PI,PI]	
		phi : float	
			latitude phi (RADIANS), phi in (-PI/2, PI/2)
		v : float
			planet-relative speed, m/s
		psi : float
			heading angle, radians
		gamma : float
			planet-relative flight-path angle, radians
		drange : float
			downrange distance measured from entry-interface

		"""

		t = self.tau*tbar
		r = rbar*self.RP
		v = vbar*self.Vref
		drange = drangebar*self.RP

		return t, r, theta, phi, v, psi, gamma, drange

	def scaleHeight(self, h, density_int):
		"""
		Returns the scale height as a function of altitude 
		for given density profile 
		
		Parameters
		----------
		h : float
			altitude at which scale height is desired
		density_int  : scipy.interpolate.interpolate.interp1d
			density interpolation function
		--
		Returns
		--
		ans : float
			scale height, meters
		"""

		# create and store an array of heights from 0 to h_skip at 
		# every 1 km
		h_array = np.linspace(0, self.h_skip, int(self.h_skip/1000.0))
		# compute the density at these altitudes using the
		# vectorized density function
		d_array = self.densityvectorized(h_array)

		# Compute the density scale height using the formula in Hamel 2006
		# AIAA. DOI = 10.2514/1.20126
		integ = cumulative_trapezoid(d_array[int(h/1000.0):], h_array[int(h/1000.0):], initial=0)[-1]
		ans = integ / (self.density(h) - self.density(self.h_skip))
		return ans

	def loadMonteCarloDensityFile2(self, atmfile, heightCol, densLowCol, densAvgCol, densHighCol, densTotalCol,
									heightInKmFlag=False):
		"""
		Loads a Monte Carlo density look up table from GRAM-Model output

		Parameters
		----------
		atmfile : str
			filename, contains mean density profile data
		heightCol : int
			column number of height values, assumes unit = meters 
			(first column = 0, second column = 1, etc.)
		densLowCol : int
			column number of the low mean density 
		densAvgCol : int
			column number of the average mean density
		densHighCol : int
			column number of the high mean desnity
		densTotalCol : int
			column number of the total (=mean + perturb.) density
		heightInKmFlag : bool, optional
			optional, set this to True if heightCol has units of km, 
			False by default

		Returns
		----------
		ATM_height : numpy.ndarray
			height array, m
		ATM_density_low : numpy.ndarray
			low density array, kg/m3
		ATM_density_avg : numpy.ndarray
			avg. density array, kg/m3
		ATM_density_high : numpy.ndarray
			high density array, kg/m3
		ATM_density_pert : numpy.ndarray
			1 sigma mean deviation from avg

		""" 
		
		# load data from textfile using np.loadtxt()
		ATM = np.loadtxt(atmfile)
		
		if heightInKmFlag is True:
			# convert heightCol from km to meters
			ATM_height = ATM[:, heightCol]*1E3
		else:
			ATM_height = ATM[:, heightCol]

		# extract data for low, avg, and mean density from file
		ATM_density_low = ATM[:, densLowCol]
		ATM_density_avg = ATM[:, densAvgCol]
		ATM_density_high = ATM[:, densHighCol]

		ATM_density_pert = ATM[:, densTotalCol] - ATM[:, densAvgCol]
		# perturb = (avg + perturb) - avg

		return ATM_height, ATM_density_low, ATM_density_avg, ATM_density_high, ATM_density_pert

	def loadMonteCarloDensityFile3(self, atmfile, heightCol, densAvgCol, densSD_percCol, densTotalCol, heightInKmFlag=False):
		"""
		Loads a Monte Carlo density look up table from GRAM-Model output,
		Utility function for EARTH-GRAM atmospheric data

		Parameters
		----------
		atmfile : str
			filename, contains mean density profile data
		heightCol : int
			column number of height values, assumes unit = meters 
			(first column = 0, second column = 1, etc.)
		densAvgCol : int
			column number of the average mean density
		densSD_percCol : int
			column number of mean density one sigma SD
		densTotalCol : int
			column number of the total (=mean + perturb.) density
		heightInKmFlag : bool, optional
			optional, set this to True if heightCol has units of km, 
			False by default

		Returns
		----------
		ATM_height : numpy.ndarray
			height array, m
		ATM_density_low : numpy.ndarray
			low density array, kg/m3
		ATM_density_avg : numpy.ndarray
			avg. density array, kg/m3
		ATM_density_high : numpy.ndarray
			high density array, kg/m3
		ATM_density_pert : numpy.ndarray
			1 sigma mean deviation from avg

		""" 
		
		# load data from textfile using np.loadtxt()
		ATM = np.loadtxt(atmfile)
		
		if heightInKmFlag is True:
			# convert heightCol from km to meters
			ATM_height = ATM[:, heightCol]*1E3
		else:
			ATM_height = ATM[:, heightCol]

		# extract data for low, avg, and mean density from file
		ATM_density_low = ATM[:, densAvgCol] - ATM[:, densAvgCol] * ATM[:, densSD_percCol]*0.01
		ATM_density_avg = ATM[:, densAvgCol]
		ATM_density_high = ATM[:, densAvgCol] + ATM[:, densAvgCol] * ATM[:, densSD_percCol]*0.01
		ATM_density_pert = ATM[:, densTotalCol] - ATM[:, densAvgCol]
		# perturb = (avg + perturb) - avg

		return ATM_height, ATM_density_low, ATM_density_avg, ATM_density_high, ATM_density_pert

	def pSigmaFunc(self, x):
		"""
		Utility function. Returns 1 if x>=0, 0.0 otherwise
		
		Parameters
		----------
		x : float
			input x

		Returns
		----------
		ans : float
			1 if x>=0, 0.0 otherwise 

		"""
		if x >= 0:
			return 1.0

		else:
			return 0.0

	def nSigmaFunc(self, x):
		"""
		Utility function. Returns 1 if x<0, 0.0 otherwise
		
		Parameters
		----------
		x : float
			input x

		Returns
		----------
		ans : float
			1 if x<0, 0.0 otherwise 

		"""
		if x < 0:
			return 1.0

		else:
			return 0.0

	def loadAtmosphereModel5(self, ATM_height, ATM_density_low, ATM_density_avg, ATM_density_high, ATM_density_pert, sigmaValue, NPOS, i):
		"""
		Read and create density_int for a single entry from a list 
		of perturbed monte carlo density profiles.
		
		Parameters
		----------
		ATM_height : numpy.ndarray
			height array, m
		ATM_density_low : numpy.ndarray
			low density array, kg/m3
		ATM_density_avg : numpy.ndarray
			avg. density array, kg/m3
		ATM_density_high : numpy.ndarray
			high density array, kg/m3
		ATM_density_pert : numpy.ndarray
			1 sigma mean deviation from avg
		sigmaValue : float
			mean desnity profile sigma deviation value
			(intended as input from a normal distribution
			with mean=0, sigma=1)
		NPOS : int
			NPOS value from GRAM model
			equals the number of positions (altitude) for which
			density value is available in look up table.
		i : int
			index of atmospheric profile to load

		Returns
		----------
		density_int : scipy.interpolate.interpolate.interp1d
			density interpolation function

		"""
		nSigma = ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)] - ATM_density_low[int((i-1)*NPOS):int(i*NPOS)]
		pSigma = ATM_density_high[int((i-1)*NPOS):int(i*NPOS)] - ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)]

		h_array = ATM_height[int((i-1)*NPOS):int(i*NPOS)]

		d_array = ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)] + pSigma*self.pSigmaFunc(sigmaValue)*sigmaValue + \
					nSigma*self.nSigmaFunc(sigmaValue)*sigmaValue + \
					ATM_density_pert[int((i-1)*NPOS):int(i*NPOS)]

		density_int = interp1d(h_array, d_array, kind='linear', fill_value=0.0, bounds_error=False)
		return density_int

	def loadAtmosphereModel6(self, ATM_height, ATM_density_low, ATM_density_avg, ATM_density_high, sigmaValue, NPOS, i):
		"""
		Read and create mean density profile for a single entry from a list 
		of perturbed monte carlo density profiles.
		
		Parameters
		----------
		ATM_height : numpy.ndarray
			height array, m
		ATM_density_low : numpy.ndarray
			low density array, kg/m3
		ATM_density_avg : numpy.ndarray
			avg. density array, kg/m3
		ATM_density_high : numpy.ndarray
			high density array, kg/m3
		sigmaValue : float
			mean desnity profile sigma deviation value
			(intended as input from a normal distribution
			with mean=0, sigma=1)
		NPOS : int
			NPOS value from GRAM model
			equals the number of positions (altitude) for which
			density value is available in look up table.
		i : int
			index of atmospheric profile to load

		Returns
		----------
		density_int : scipy.interpolate.interpolate.interp1d
			density interpolation function

		"""
		nSigma = ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)] - ATM_density_low[int((i-1)*NPOS):int(i*NPOS)]
		pSigma = ATM_density_high[int((i-1)*NPOS):int(i*NPOS)] - ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)]

		h_array = ATM_height[int((i-1)*NPOS):int(i*NPOS)]

		d_array = ATM_density_avg[int((i-1)*NPOS):int(i*NPOS)] + pSigma*self.pSigmaFunc(sigmaValue)*sigmaValue +\
					nSigma*self.nSigmaFunc(sigmaValue)*sigmaValue

		density_int = interp1d(h_array, d_array, kind='linear', fill_value=0.0, bounds_error=False)
		return density_int



import numpy as np
from mayavi import mlab

from AMAT.approach import Approach
from AMAT.planet import Planet
from AMAT.vehicle import Vehicle
from AMAT.orbiter import Orbiter, PropulsiveOrbiter
from astropy.time import Time
from AMAT.arrival import Arrival

arrival = Arrival()
arrival.set_vinf_vec_from_lambert_arc(lastFlybyPlanet='EARTH',
									  arrivalPlanet='VENUS',
									  lastFlybyDate=Time("2010-05-10 00:00:00", scale='tdb'),
									  arrivalDate=Time("2010-12-06 00:00:00", scale='tdb'),
									  ephem_file='../../../spice-data/de432s.bsp')


probe = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+200)*1e3, psi=np.pi)


space = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+400)*1e3, psi=np.pi)


north_pole_bi_vec = probe.ICRF_to_BI(arrival.north_pole)





orbiter = PropulsiveOrbiter(approach=probe, apoapsis_alt_km=2e3)
orbiter2 = PropulsiveOrbiter(approach=space, apoapsis_alt_km=150e3)


u = np.linspace(0, 2 * np.pi, 100)
v = np.linspace(0, np.pi, 100)
x = 1 * np.outer(np.cos(u), np.sin(v))
y = 1 * np.outer(np.sin(u), np.sin(v))
z = 1 * np.outer(np.ones(np.size(u)), np.cos(v))


x1 = 1.02478929103*np.outer(np.cos(u), np.sin(v))
y1 = 1.024789291032*np.outer(np.sin(u), np.sin(v))
z1 = 1.02478929103*np.outer(np.ones(np.size(u)), np.cos(v))


x_ring_1 = 1.1*np.cos(u)
y_ring_1 = 1.1*np.sin(u)
z_ring_1 = 0.0*np.cos(u)

x_ring_2 = 1.2*np.cos(u)
y_ring_2 = 1.2*np.sin(u)
z_ring_2 = 0.0*np.cos(u)

mlab.figure(bgcolor=(0,0,0))
s1 = mlab.mesh(x, y, z, color=(0.34,0.33,0.33))
s2 = mlab.mesh(x1, y1, z1, color=(0.34, 0.33,0.33), opacity=0.3)
r1 = mlab.plot3d(x_ring_1, y_ring_1, z_ring_1, color=(1,1,1), line_width=1, tube_radius=None)



p4 = mlab.plot3d(orbiter.x_orbit_arr,
				 orbiter.y_orbit_arr,
				 orbiter.z_orbit_arr, color=(0, 1, 0), line_width=3, tube_radius=None)

p5 = mlab.plot3d(orbiter2.x_orbit_arr,
				 orbiter2.y_orbit_arr,
				 orbiter2.z_orbit_arr, color=(1, 0, 1), line_width=3, tube_radius=None)


mlab.plot3d([0, 1.05 * north_pole_bi_vec[0]],
                   [0, 1.05 * north_pole_bi_vec[1]],
                   [0, 1.05 * north_pole_bi_vec[2]])

mlab.show()



"""
test_vehicle.py

Tests for Vehicle class

"""

import unittest
import numpy as np


try:
	from AMAT.planet import Planet
except ModuleNotFoundError:
	raise ModuleNotFoundError("Cannot import Planet from AMAT.planet")

try:
	from AMAT.vehicle import Vehicle
except ModuleNotFoundError:
	raise ModuleNotFoundError("Cannot import Vehicle from AMAT.vehicle")


class TestBallisticEntries(unittest.TestCase):
	"""
	Create vehicles and propagate a few ballistic planetary
	entry vehicle trajectories.
	"""

	def test_create_vehicle_pioneer_venus(self):
		"""
		Creates a test vehicle flying in Venus atmosphere
		"""
		planet1 = Planet("VENUS")
		# load atmosphere models
		try:
			planet1.loadAtmosphereModel('atmdata/Venus/venus-gram-avg.dat', 0, 1, 2, 3)
		except OSError:
			raise OSError("File not found. Check file path/name, and make sure file is present.")

		# pv-large probe
		vehicle1 = Vehicle('PV-Large', 316, 188, 0.00, 1.59, 0.0, 0.36, planet1)

		vehicle1.setInitialState(180.0, 0.0, 0.0, 11.54, 0.0, -32.4, 0.0, 0.0)
		vehicle1.setSolverParams(1E-6)
		vehicle1.propogateEntry(3600.0, 0.1, 0.0)

		# Check computed peak heating and peak g-load values
		self.assertAlmostEqual(max(vehicle1.q_stag_total), 4935, delta=100)
		self.assertAlmostEqual(max(vehicle1.acc_net_g), 287, delta=10)
		# Check vehicle.exitflag == -1 (indicates vehicle reached "trap" altitude)
		self.assertEqual(vehicle1.exitflag, -1.0)

	def test_create_vehicle_stardust_earth(self):
		planet2 = Planet("EARTH")
		planet2.h_skip = 125.0E3

		try:
			planet2.loadAtmosphereModel('atmdata/Earth/earth-gram-avg.dat', 0 , 1 ,2, 3)
		except OSError:
			raise OSError("File not found. Check file path/name, and make sure file is present.")

		vehicle2 = Vehicle('Stardust', 45.8, 60.0, 0.00, 0.52, 0.0, 0.23, planet2)  # stardust
		vehicle2.setInitialState(125.0, 0.0, 0.0, 12.6, 0.0, -8.2, 0.0, 0.0)
		vehicle2.setSolverParams(1E-6)
		vehicle2.propogateEntry(1800.0, 0.1, 0.0)

		self.assertAlmostEqual(max(vehicle2.q_stag_total), 824, delta=100)
		self.assertAlmostEqual(max(vehicle2.acc_net_g), 32, delta=4)
		self.assertEqual(vehicle2.exitflag, -1.0)

	def test_create_vehicle_curiosity_mars(self):
		planet3 = Planet("MARS")
		planet3.h_skip = 126.0E3
		planet3.h_trap = 2.0E3

		try:
			planet3.loadAtmosphereModel('atmdata/Mars/mars-gram-avg.dat', 0 , 1 ,2, 3)
		except OSError:
			raise OSError("File not found. Check file path/name, and make sure file is present.")


		vehicle3=Vehicle('Curiosity', 3257.0, 146.0, 0.00, np.pi*4.5**2.0, 0.0, 1.125, planet3) # curiosity
		vehicle3.setInitialState(125.0,0.0,0.0,6.08,0.0,-15.48,0.0,0.0)
		vehicle3.setSolverParams(1E-6)
		vehicle3.propogateEntry(1800.0,0.1,0.0)

		self.assertAlmostEqual(max(vehicle3.q_stag_total), 81, delta=10)
		self.assertAlmostEqual(max(vehicle3.acc_net_g), 13, delta=3)
		self.assertEqual(vehicle3.exitflag, -1.0)



if __name__ == '__main__':
	unittest.main()

		








"""
test_manever.py

Tests for maneuver.py

"""

import numpy as np


try:
    from AMAT.maneuver import ProbeOrbiterDeflection, ProbeProbeDeflection, OrbiterOrbiterDeflection
except ModuleNotFoundError:
    raise ModuleNotFoundError("Cannot import classes from AMAT.maneuver")


class Test_ProbeOrbiterDeflection_Neptune:

    deflection = ProbeOrbiterDeflection("NEPTUNE", v_inf_vec_icrf_kms=np.array([17.78952518,  8.62038536,  3.15801163]),
                                       rp_probe=(24622+400)*1e3,  psi_probe=3*np.pi/2, h_EI_probe=1000e3,
                                       rp_space=(24622+4000)*1e3, psi_space=np.pi/2,
                                       r_dv_rp=1000)

    def test_theta_star_dv_probe(self):
        ans = self.deflection.theta_star_dv_probe
        rdv = self.deflection.probe.r_mag_bi(ans)/self.deflection.planetObj.RP
        assert abs(ans + 1.9866386) < 1e-6
        assert abs(rdv - self.deflection.r_dv_rp) < 1e-6

    def test_delta_theta_star_probe(self):
        assert abs(self.deflection.delta_theta_star_probe - 1.9866386) < 1e-6


    def test_v_vec_dv_maneuver(self):
        assert abs(self.deflection.dv_maneuver_mag - 65.7302) < 1e-2
        assert abs(self.deflection.TOF_probe - 14.17563) < 1e-2
        assert abs(self.deflection.TOF_space - 14.17647) < 1e-2


class Test_ProbeProbeDeflection_Neptune:

    deflection = ProbeProbeDeflection( "NEPTUNE",
                                       v_inf_vec_icrf_kms=np.array([17.78952518,  8.62038536,  3.15801163]),
                                       rp_probe1=(24622+400)*1e3,  psi_probe1=3*np.pi/2, h_EI_probe1=1000e3,
                                       rp_probe2=(24622+400)*1e3,  psi_probe2=np.pi/2,  h_EI_probe2=1000e3,
                                       r_dv_rp=4000)


    def test_v_vec_dv_maneuver(self):
        assert abs(self.deflection.dv_maneuver_mag - 15.6355338) < 1e-6
        assert abs(self.deflection.TOF_probe1 - 56.8684222) < 1e-6
        assert abs(self.deflection.TOF_probe2 - 56.8684222) < 1e-6
        assert abs(self.deflection.probe1.gamma_entry_inertial*180/np.pi - -10.5025297) < 1e-6
        assert abs(self.deflection.probe2.gamma_entry_inertial*180/np.pi - -10.5025297) < 1e-6
        assert abs(self.deflection.probe1.heading_entry_atm*180/np.pi - 14.9106075) < 1e-6
        assert abs(self.deflection.probe2.heading_entry_atm*180/np.pi - 167.5494031) < 1e-6


class Test_OrbiterOrbiterDeflection_Neptune:

    deflection = OrbiterOrbiterDeflection( "NEPTUNE",
                                       v_inf_vec_icrf_kms=np.array([17.78952518,  8.62038536,  3.15801163]),
                                       rp_space1=(24622+4000)*1e3,  psi_space1=3*np.pi/2,
                                       rp_space2=(24622+4000)*1e3,  psi_space2=np.pi/2,
                                       r_dv_rp=1000)


    def test_v_vec_dv_maneuver(self):
        assert abs(self.deflection.dv_maneuver_mag - 68.9076850) < 1e-6
        assert abs(self.deflection.TOF_space1 - 14.17643992231) < 1e-6
        assert abs(self.deflection.TOF_space2 - 14.17647834316) < 1e-6



import numpy as np
from scipy.interpolate import interp1d

class Launcher:
	"""
	The Launcher class is used to estimate launch vehicle performance.
	
	Attributes
	----------
	launcherID : str
		String identifier for launch vehicle
	datafile : CSV file
		CSV file containing C3,launch mass (kg)
	kind : str
			type of interpolation to use. Defaults to 'linear'
	f : scipy.interpolate.interp1d
		interpolation function for launch mass at a specified C3
	"""

	def __init__(self, launcherID, datafile, kind='linear'):
		"""
		Initializes the launcher class.
		
		Parameters
		----------
		launcherID : str
			identifier for the launch vehicle
		datafile : file
			CSV file containing the C3,launch mass data
		kind : str
			type of interpolation to use. Defaults to 'linear'.

		"""

		self.ID = launcherID
		self.XY = np.loadtxt(datafile, delimiter=',')
		self.f = interp1d(self.XY[:, 0], self.XY[:, 1], kind=kind, fill_value=0, bounds_error=False)

	def launchMass(self, C3):
		"""
		Returns the launch capability of the vehicle for a 
		specified C3 array.

		Parameters
		----------
		C3 : float
			launch C3, km2/s2

		Returns
		--------
		mass : float
			launch mass capability, kg

		"""
		mass = self.f(C3)
		return mass











		



from mayavi import mlab
import numpy as np
from tvtk.tools import visual
from AMAT.approach import Approach
from astropy.time import Time
from AMAT.arrival import Arrival
from AMAT.vehicle import Vehicle

def Arrow_From_A_to_B(x1, y1, z1, x2, y2, z2):
    ar1 = visual.arrow(x=x1, y=y1, z=z1)
    ar1.length_cone = 0.4

    arrow_length = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)
    ar1.actor.scale = [arrow_length, arrow_length, arrow_length]
    ar1.pos = ar1.pos / arrow_length
    ar1.axis = [x2 - x1, y2 - y1, z2 - z1]
    return ar1

arrival = Arrival()
arrival.set_vinf_vec_from_lambert_arc(lastFlybyPlanet='EARTH',
									  arrivalPlanet='VENUS',
									  lastFlybyDate=Time("2010-05-10 00:00:00", scale='tdb'),
									  arrivalDate=Time("2010-12-06 00:00:00", scale='tdb'),
									  ephem_file='../../../spice-data/de432s.bsp')


probe1 = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+103.85)*1e3, psi=1.05988*np.pi,
                            is_entrySystem=True, h_EI=150e3)

probe2 = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+103.85)*1e3, psi=1.03*np.pi,
                            is_entrySystem=True, h_EI=150e3)

probe3 = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+103.85)*1e3, psi=np.pi,
                            is_entrySystem=True, h_EI=150e3)

probe4 = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+103.85)*1e3, psi=0.97*np.pi,
                            is_entrySystem=True, h_EI=150e3)

probe5 = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec,
                            rp=(6051.8+103.85)*1e3, psi=0.94*np.pi,
                            is_entrySystem=True, h_EI=150e3)

space = Approach("VENUS", v_inf_vec_icrf_kms=arrival.v_inf_vec, rp=(6051.8 + 400) * 1e3, psi=1.0*np.pi)



theta_star_arr_space = np.linspace(-2.00, 0.0, 101)
pos_vec_bi_arr_space = space.pos_vec_bi(theta_star_arr_space)/6051.8e3

x_arr_space = pos_vec_bi_arr_space[0][:]
y_arr_space = pos_vec_bi_arr_space[1][:]
z_arr_space = pos_vec_bi_arr_space[2][:]

north_pole_bi_vec = probe1.ICRF_to_BI(arrival.north_pole)

theta_star_arr_probe1 = np.linspace(-2, probe1.theta_star_entry, 101)
pos_vec_bi_arr_probe1 = probe1.pos_vec_bi(theta_star_arr_probe1)/6051.8e3

theta_star_arr_probe2 = np.linspace(-2, probe2.theta_star_entry, 101)
pos_vec_bi_arr_probe2 = probe2.pos_vec_bi(theta_star_arr_probe2)/6051.8e3

theta_star_arr_probe3 = np.linspace(-2, probe3.theta_star_entry, 101)
pos_vec_bi_arr_probe3 = probe3.pos_vec_bi(theta_star_arr_probe3)/6051.8e3

theta_star_arr_probe4 = np.linspace(-2, probe4.theta_star_entry, 101)
pos_vec_bi_arr_probe4 = probe4.pos_vec_bi(theta_star_arr_probe4)/6051.8e3

theta_star_arr_probe5 = np.linspace(-2, probe5.theta_star_entry, 101)
pos_vec_bi_arr_probe5 = probe5.pos_vec_bi(theta_star_arr_probe5)/6051.8e3


x_arr_probe1 = pos_vec_bi_arr_probe1[0][:]
y_arr_probe1 = pos_vec_bi_arr_probe1[1][:]
z_arr_probe1 = pos_vec_bi_arr_probe1[2][:]

x_arr_probe2 = pos_vec_bi_arr_probe2[0][:]
y_arr_probe2 = pos_vec_bi_arr_probe2[1][:]
z_arr_probe2 = pos_vec_bi_arr_probe2[2][:]

x_arr_probe3 = pos_vec_bi_arr_probe3[0][:]
y_arr_probe3 = pos_vec_bi_arr_probe3[1][:]
z_arr_probe3 = pos_vec_bi_arr_probe3[2][:]

x_arr_probe4 = pos_vec_bi_arr_probe4[0][:]
y_arr_probe4 = pos_vec_bi_arr_probe4[1][:]
z_arr_probe4 = pos_vec_bi_arr_probe4[2][:]

x_arr_probe5 = pos_vec_bi_arr_probe5[0][:]
y_arr_probe5 = pos_vec_bi_arr_probe5[1][:]
z_arr_probe5 = pos_vec_bi_arr_probe5[2][:]



u = np.linspace(0, 2 * np.pi, 100)
v = np.linspace(0, np.pi, 100)
x = 1*np.outer(np.cos(u), np.sin(v))
y = 1*np.outer(np.sin(u), np.sin(v))
z = 1*np.outer(np.ones(np.size(u)), np.cos(v))

x1 = 1.02478929103*np.outer(np.cos(u), np.sin(v))
y1 = 1.024789291032*np.outer(np.sin(u), np.sin(v))
z1 = 1.02478929103*np.outer(np.ones(np.size(u)), np.cos(v))



x_ring_1 = 1.1*np.cos(u)
y_ring_1 = 1.1*np.sin(u)
z_ring_1 = 0.0*np.cos(u)

x_ring_2 = 1.2*np.cos(u)
y_ring_2 = 1.2*np.sin(u)
z_ring_2 = 0.0*np.cos(u)

mlab.figure(bgcolor=(0,0,0))
s1 = mlab.mesh(x, y, z, color=(0.34,0.33,0.33))
s2 = mlab.mesh(x1, y1, z1, color=(0.34, 0.33,0.33), opacity=0.3)
r1 = mlab.plot3d(x_ring_1, y_ring_1, z_ring_1, color=(1,1,1), line_width=1, tube_radius=None)
#r2 = mlab.plot3d(x_ring_2, y_ring_2, z_ring_2, color=(1,1,1), line_width=1, tube_radius=None)

p1 = mlab.plot3d(x_arr_probe1, y_arr_probe1, z_arr_probe1, color=(0,1,0), line_width=3, tube_radius=None)
p2 = mlab.plot3d(x_arr_probe2, y_arr_probe2, z_arr_probe2, color=(0,1,0), line_width=3, tube_radius=None)
p3 = mlab.plot3d(x_arr_probe3, y_arr_probe3, z_arr_probe3, color=(0,1,0), line_width=3, tube_radius=None)
p4 = mlab.plot3d(x_arr_probe4, y_arr_probe4, z_arr_probe4, color=(0,1,0), line_width=3, tube_radius=None)
p5 = mlab.plot3d(x_arr_probe5, y_arr_probe5, z_arr_probe5, color=(0,1,0), line_width=3, tube_radius=None)
#p2 = mlab.plot3d(x_arr_space, y_arr_space, z_arr_space, color=(0,0,1), line_width=3, tube_radius=None)
o1 = mlab.plot3d(x_arr_space, y_arr_space, z_arr_space, color=(1,0,1), line_width=3, tube_radius=None)


mlab.plot3d([0, 1.2 * north_pole_bi_vec[0]],
                   [0, 1.2 * north_pole_bi_vec[1]],
                   [0, 1.2 * north_pole_bi_vec[2]])

# mlab.plot3d([0, 1.5 * probe1.S_vec_bi_unit[0]],
#                    [0, 1.5 * probe1.S_vec_bi_unit[1]],
#                    [0, 1.5 * probe1.S_vec_bi_unit[2]], color=(0, 1, 0))
#
# mlab.plot3d([0, 1.5 * probe1.R_vec_bi_unit[0]],
#                    [0, 1.5 * probe1.R_vec_bi_unit[1]],
#                    [0, 1.5 * probe1.R_vec_bi_unit[2]], color=(1, 0, 0))
#
# mlab.plot3d([0, 1.5 * probe1.T_vec_bi_unit[0]],
#                    [0, 1.5 * probe1.T_vec_bi_unit[1]],
#                    [0, 1.5 * probe1.T_vec_bi_unit[2]], color=(0, 0, 1))


mlab.show()



"""
test_girija2020.py

Compare results from Girija et al. [2020, JSR]

"""

import unittest
import numpy as np

try:
    from AMAT.planet import Planet
except ModuleNotFoundError:
    raise ModuleNotFoundError("Cannot import Planet from AMAT.planet")

try:
	from AMAT.vehicle import Vehicle
except ModuleNotFoundError:
	raise ModuleNotFoundError("Cannot import Vehicle from AMAT.vehicle")

try:
    from AMAT.launcher import Launcher
except ModuleNotFoundError:
    raise ModuleNotFoundError("Cannot import Launcher from AMAT.Launcher")


class Girija2020(unittest.TestCase):

	def test_aerocapture_vehicle(self):

		planet = Planet("NEPTUNE")

		try:
			planet.loadAtmosphereModel('atmdata/Neptune/neptune-gram-avg.dat', 0 , 7 ,6, 5 , heightInKmFlag=True)
		except OSError:
			raise OSError("File not found. Check file path/name, and make sure file is present.")

		# select L/D = 0.4 and Vinf_arrival = 20 km/s from feasibility chart

		# C3 of selected trajectory = 111 km2/s2
		# selected launch vehicle = SLS-Block-1B with kick stage

		# compute launch vehicle capability
		launcher = Launcher( launcherID='sls-block-1B-with-kick',
							 datafile='launcher-data/sls-block-1B-with-kick.csv')
		launch_mass = launcher.launchMass(111.0)
		assert abs(launch_mass - 6600) < 100

		# vehicle and entry state description from Girija et al. [JSR, 2021]
		vehicle = Vehicle('Trident', 1000.0, 200.0, 0.40, 3.1416, 0.0, 1.00, planet)
		vehicle.setInitialState(1000.0, 0.0, 0.0, 28.00, 0.0,-10.00, 0.0, 0.0)
		vehicle.setSolverParams(1E-6)

		print("Computing aerocapture corridor bounds from Girija et al. [JSR, 2020]...")
		overShootLimit, exitflag_os  = vehicle.findOverShootLimit (2400.0, 1.0, -20.0, -4.0, 1E-2, 400.0e3)
		underShootLimit,exitflag_us  = vehicle.findUnderShootLimit(2400.0, 1.0, -20.0, -4.0, 1E-2, 400.0e3)

		# NOTE: This result does not include the effect of planetary rotation on
		# entry speed. Use findOverShootLimit2 and findUnderShootLimit2 to include these
		# effects on the corridor location.

		# check exitflag_os, exitflag_us = 1 (indicates AMAT found a solution)
		self.assertEqual(exitflag_os, 1)
		self.assertEqual(exitflag_us, 1)

		# compare with values from Fig. 2 
		self.assertAlmostEqual(overShootLimit,  -12.7393, places=2)
		self.assertAlmostEqual(underShootLimit, -13.7081, places=2)


if __name__ == '__main__':
	unittest.main()




